#!/usr/bin/env python3

import _setup_env  # noqa
import argparse
import time
import math
import os
import torch
import better_exchook

import data


my_dir = os.path.dirname(os.path.abspath(__file__))


def main():

  parser = argparse.ArgumentParser(description='PyTorch Wikitext-2 RNN/LSTM/GRU/Transformer Language Model')
  parser.add_argument('--data', type=str, default=f'{my_dir}/data/wikitext-2',
                      help='location of the data corpus')
  parser.add_argument('--model', type=str, default='LSTM',
                      help='type of recurrent net (RNN_TANH, RNN_RELU, LSTM, GRU, Transformer)')
  parser.add_argument('--emsize', type=int, default=200,
                      help='size of word embeddings')
  parser.add_argument('--nhid', type=int, default=200,
                      help='number of hidden units per layer')
  parser.add_argument('--nlayers', type=int, default=2,
                      help='number of layers')
  parser.add_argument('--lr', type=float, default=20,
                      help='initial learning rate')
  parser.add_argument('--clip', type=float, default=0.25,
                      help='gradient clipping')
  parser.add_argument('--epochs', type=int, default=40,
                      help='upper epoch limit')
  parser.add_argument('--batch_size', type=int, default=20, metavar='N',
                      help='batch size')
  parser.add_argument('--bptt', type=int, default=35,
                      help='sequence length')
  parser.add_argument('--dropout', type=float, default=0.2,
                      help='dropout applied to layers (0 = no dropout)')
  parser.add_argument('--tied', action='store_true',
                      help='tie the word embedding and softmax weights')
  parser.add_argument('--seed', type=int, default=1111,
                      help='random seed')
  parser.add_argument('--cuda', action='store_true',
                      help='use CUDA')
  parser.add_argument('--log-interval', type=int, default=200, metavar='N',
                      help='report interval')
  parser.add_argument('--save', type=str, default='model.pt',
                      help='path to save the final model')
  parser.add_argument('--onnx-export', type=str, default='',
                      help='path to export the final model in onnx format')

  parser.add_argument('--nhead', type=int, default=2,
                      help='the number of heads in the encoder/decoder of the transformer model')
  parser.add_argument('--dry-run', action='store_true',
                      help='verify the code and the model')

  args = parser.parse_args()

  # Set the random seed manually for reproducibility.
  torch.manual_seed(args.seed)
  if torch.cuda.is_available():
      if not args.cuda:
          print("WARNING: You have a CUDA device, so you should probably run with --cuda")

  device = torch.device("cuda" if args.cuda else "cpu")

  ###############################################################################
  # Load data
  ###############################################################################

  corpus = data.Corpus(args.data)

  # Starting from sequential data, batchify arranges the dataset into columns.
  # For instance, with the alphabet as the sequence and batch size 4, we'd get
  # ┌ a g m s ┐
  # │ b h n t │
  # │ c i o u │
  # │ d j p v │
  # │ e k q w │
  # └ f l r x ┘.
  # These columns are treated as independent by the model, which means that the
  # dependence of e. g. 'g' on 'f' can not be learned, but allows more efficient
  # batch processing.

  def batchify(data, bsz):
    # Work out how cleanly we can divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    return data.to(device)

  eval_batch_size = 10
  train_data = batchify(corpus.train, args.batch_size)
  val_data = batchify(corpus.valid, eval_batch_size)
  test_data = batchify(corpus.test, eval_batch_size)

  def get_batch(source, i):
    seq_len = min(args.bptt, len(source) - 1 - i)
    data = source[i: i +seq_len]
    target = source[ i +1: i + 1 +seq_len].view(-1)
    return data, target

  ntokens = len(corpus.dictionary)

  def model_func(wrapped_import, inputs):
    ###############################################################################
    # Build the model
    ###############################################################################
    if wrapped_import:
      nn = wrapped_import("torch.nn")
      model = wrapped_import("model")
    else:
      from torch import nn
      import model

    if args.model == 'Transformer':
      net = model.TransformerModel(ntokens, args.emsize, args.nhead, args.nhid, args.nlayers, args.dropout)
    else:
      net = model.RNNModel(args.model, ntokens, args.emsize, args.nhid, args.nlayers, args.dropout, args.tied)

    net.eval()  # for verification, need no random elements (e.g. dropout)
    # criterion = nn.NLLLoss()

    if args.model != 'Transformer':
      hidden = net.init_hidden(bsz=inputs.shape[1])
    else:
      hidden = None
    with torch.no_grad():
      if args.model == 'Transformer':
        output = net(inputs)
        output = output.view(-1, ntokens)
      else:
        output, hidden = net(inputs, hidden)

      return output

  data_, targets = get_batch(train_data, i=0)

  from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn
  verify_torch_and_convert_to_returnn(
    model_func,
    inputs=data_.detach().cpu().numpy(),
    inputs_data_kwargs={"shape": (None,), "sparse": True, "dim": ntokens, "batch_dim_axis": 1})


if __name__ == '__main__':
  better_exchook.install()
  main()




from .object import make_wrapped_object, WrappedObject
from .class_ import make_wrapped_class, WrappedClassBase
from .function import make_wrapped_function
from .module import WrappedModule, WrappedSourceModule, WrappedIndirectModule




import types
from ... import log
from .object import WrappedObject
from .utils import copy_attribs_qualname_and_co
from ..context import WrapCtx


def make_wrapped_function(func, name: str, ctx: WrapCtx):
  from ..wrap import wrap, unwrap

  def _call(*args, **kwargs):
    if log.Verbosity >= 3:
      log.unique_print("*** func call %s(...)" % name)
    res = func(*unwrap(args), **unwrap(kwargs))
    res = wrap(res, name="%s(...)" % name, ctx=ctx)
    return res

  copy_attribs_qualname_and_co(_call, func, ctx=ctx)
  if isinstance(func, _FuncType):
    return _call

  # This is maybe slightly unconventional / non-straightforward.
  # We construct it such to make PyTorch _add_docstr happy,
  # which expects certain types.
  class WrappedFunc(WrappedObject):
    def __new__(cls, *args, **kwargs):
      if log.Verbosity >= 3:
        log.unique_print("*** func call %s(...)" % name)
      return _call(*args, **kwargs)

  copy_attribs_qualname_and_co(WrappedFunc, func, ctx=ctx)
  return WrappedFunc


# https://youtrack.jetbrains.com/issue/PY-45382
# noinspection PyTypeChecker
_FuncType = types.FunctionType  # type: type




"""
This wraps PyTorch ``torch.nn.modules.module``.
"""


from __future__ import annotations
import tensorflow as tf
from tensorflow.python.util import nest
import numpy
from collections import OrderedDict
from typing import Optional, Callable, TypeVar, Iterator, Tuple, List, Set, Union, Dict, Any, Collection
import types
import itertools
from ..parameter import Parameter
from ...tensor import Tensor
from ...autograd import no_grad
from ...utils.hooks import RemovableHandle
from ....naming import Naming, CallEntry, TensorEntry
from returnn.tf.layers.basic import LayerBase, SubnetworkLayer
from returnn.tf.util.data import Data, DimensionTag

# See https://mypy.readthedocs.io/en/latest/generics.html#generic-methods-and-generic-self for the use
# of `T` to annotate `self`. Many methods of `Module` return `self` and we want those return values to be
# the type of the subclass, not the looser type of `Module`.
T = TypeVar('T', bound='Module')


class Module:
  """
  Base class.

  Note::

  When porting over a PyTorch module to here, you should keep the module mostly as-is,
  i.e. mostly use the original code, and keep all parameters, buffers, etc.
  You can even keep the `forward` function as long as all other modules/functions it uses are wrapped.
  Or you remove the `forward` function, and implement `create_returnn_layer_dict`.

  Low-level modules which would wrap directly to a corresponding RETURNN layer,
  should do that, i.e. have no `forward`, but implement `create_returnn_layer_dict`.
  (If the original `forward` returns not a single tensor, but a tuple, or some other nested structure,
   override `make_structured_returnn_output`.)

  Other modules should work just as-is.
  I.e. this can be used as base class for external PyTorch code.

  This would wrap all the standard PyTorch modules (e.g. torch.nn.Conv1d)
  (in that case, `is_original_torch_module = True`).
  This would also be used to implement new modules
  which are needed to wrap other functions (e.g. torch.nn.functional)
  (in that case, `is_original_torch_module = False`).
  """
  # All derived classes here, which exist in PyTorch as well (e.g. torch.nn.Conv1d etc),
  # should have this set to True.
  is_original_torch_module: bool = True
  _forward_feed_dict_deps: bool = False  # for internal usage

  _wrapped_class_cache = {}  # cls -> WrappedClass

  # Need to overwrite to wrap __init__ to correctly set context.
  def __new__(cls, *args, **kwargs):
    if cls not in cls._wrapped_class_cache:
      class WrappedClass(cls):
        def __init__(self, *args, **kwargs):
          self.__class__ = cls  # we don't need this wrapper class anymore
          with Naming.get_instance().push_module_creation(self):
            cls.__init__(self, *args, **kwargs)
      WrappedClass.__name__ = cls.__name__
      WrappedClass.__qualname__ = cls.__qualname__
      WrappedClass.__module__ = cls.__module__
      wrapped_cls = WrappedClass
      cls._wrapped_class_cache[cls] = wrapped_cls
    else:
      wrapped_cls = cls._wrapped_class_cache[cls]
    return super(Module, cls).__new__(wrapped_cls)

  def __init__(self):
    self._parameters = OrderedDict()  # type: OrderedDict[str, Parameter]
    self._modules = OrderedDict()  # type: OrderedDict[str, Optional[Module]]
    self._buffers = OrderedDict()  # type: OrderedDict[str, Tensor]
    self._non_persistent_buffers_set = set()
    self._forward_pre_hooks = OrderedDict()

  def __repr__(self):
    return f"<{self.__class__.__name__}>"

  def get_returnn_name(self) -> str:
    return self.__class__.__name__

  # Overwrite to catch all access, to be able to wrap member functions.
  def __getattribute__(self, item: str):
    if item in {"__getattr__", "__setattr__", "__delattr__", "__dir__", "__dict__", "__class__"}:
      # Fast path, and no need for wrapping.
      return super(Module, self).__getattribute__(item)
    if hasattr(Module, item) and getattr(self.__class__, item) is getattr(Module, item):
      # Some member of the base class.
      # No wrapping needed.
      return super(Module, self).__getattribute__(item)
    obj = super(Module, self).__getattribute__(item)
    if isinstance(obj, types.MethodType):
      def wrapped_func(*args, **kwargs):
        with Naming.get_instance().push_module_context(self):
          return obj(*args, **kwargs)
      wrapped_func.__name__ = obj.__name__
      wrapped_func.__qualname__ = obj.__qualname__
      return wrapped_func
    return obj

  def __getattr__(self, item):
    if item in {"_parameters", "_modules", "_buffers"}:
      raise AttributeError(f"Module.__init__ not yet called, attrib {item!r} not present")
    if item in self._parameters:
      return self._parameters[item]
    if item in self._buffers:
      return self._buffers[item]
    if item in self._modules:
      return self._modules[item]
    raise AttributeError(f"No attrib {item!r}")

  def __setattr__(self, name, value):
    def remove_from(*dicts_or_sets):
      for d in dicts_or_sets:
        if name in d:
          if isinstance(d, dict):
            del d[name]
          else:
            d.discard(name)

    params = self.__dict__.get('_parameters')
    if isinstance(value, Parameter):
      if params is None:
        raise AttributeError("cannot assign parameters before Module.__init__() call")
      remove_from(self.__dict__, self._buffers, self._modules, self._non_persistent_buffers_set)
      self.register_parameter(name, value)
      return

    if params is not None and name in params:
      if value is not None:
        raise TypeError(f"cannot assign {type(value)} as parameter {name!r} (torch.nn.Parameter or None expected)")
      self.register_parameter(name, value)
      return

    modules = self.__dict__.get('_modules')
    if isinstance(value, Module):
      if modules is None:
        raise AttributeError("cannot assign module before Module.__init__() call")
      remove_from(self.__dict__, self._parameters, self._buffers, self._non_persistent_buffers_set)
      self.add_module(name, value)
      return

    if modules is not None and name in modules:
      if value is not None:
        raise TypeError(f"cannot assign {type(value)} as child module {name!r} (torch.nn.Module or None expected)")
      modules[name] = value
      return

    buffers = self.__dict__.get('_buffers')
    if buffers is not None and name in buffers:
      if value is not None and not isinstance(value, Tensor):
        raise TypeError(f"cannot assign {type(value)} as buffer {name!r} (torch.Tensor or None expected)")
      buffers[name] = value
      if value is not None:
        Naming.get_instance().register_module_child_attr(self, name, value)
      return

    object.__setattr__(self, name, value)
    if isinstance(value, Tensor):
      Naming.get_instance().register_module_child_attr(self, name, value)

  def __delattr__(self, item):
    if item in self._parameters:
      del self._parameters[item]
      return
    if item in self._buffers:
      del self._buffers[item]
      return
    if item in self._modules:
      del self._modules[item]
      return
    super(Module, self).__delattr__(item)

  def __dir__(self):
    module_attrs = dir(self.__class__)
    attrs = list(self.__dict__.keys())
    parameters = list(self._parameters.keys())
    modules = list(self._modules.keys())
    buffers = list(self._buffers.keys())
    keys = module_attrs + attrs + parameters + modules + buffers
    keys = [key for key in keys if not key[0].isdigit()]  # Eliminate attrs that are not legal Python variable names
    return sorted(keys)

  def children(self) -> Iterator['Module']:
    for name, module in self.named_children():
      yield module

  def named_children(self) -> Iterator[Tuple[str, 'Module']]:
    memo = set()
    for name, module in self._modules.items():
      if module is not None and module not in memo:
        memo.add(module)
        yield name, module

  def add_module(self, name: str, module: Optional['Module']) -> None:
    if not isinstance(module, Module) and module is not None:
      raise TypeError(f"{type(module)} is not a Module subclass")
    elif not isinstance(name, str):
      raise TypeError(f"module name should be a string. Got {type(name)}")
    elif hasattr(self, name) and name not in self._modules:
      raise KeyError(f"attribute {name!r} already exists")
    elif '.' in name:
      raise KeyError("module name can't contain \".\"")
    elif name == '':
      raise KeyError("module name can't be empty string \"\"")
    self._modules[name] = module
    if module is not None:
      Naming.get_instance().register_module_child_attr(self, name, module)

  def register_parameter(self, name: str, param: Optional[Parameter]) -> None:
    if '_parameters' not in self.__dict__:
      raise AttributeError("cannot assign parameter before Module.__init__() call")
    elif not isinstance(name, str):
      raise TypeError(f"parameter name should be a string. Got {type(name)}")
    elif '.' in name:
      raise KeyError("parameter name can't contain \".\"")
    elif name == '':
      raise KeyError("parameter name can't be empty string \"\"")
    elif hasattr(self, name) and name not in self._parameters:
      raise KeyError("attribute '{}' already exists".format(name))
    self._parameters[name] = param
    if param is not None:
      Naming.get_instance().register_module_child_attr(self, name, param)

  def register_buffer(self, name: str, tensor: Optional[Tensor], persistent: bool = True) -> None:
    if '_buffers' not in self.__dict__:
      raise AttributeError("cannot assign buffer before Module.__init__() call")
    elif not isinstance(name, str):
      raise TypeError(f"buffer name should be a string. Got {type(name)}")
    elif '.' in name:
      raise KeyError("buffer name can't contain \".\"")
    elif name == '':
      raise KeyError("buffer name can't be empty string \"\"")
    elif hasattr(self, name) and name not in self._buffers:
      raise KeyError("attribute '{}' already exists".format(name))
    elif tensor is not None and not isinstance(tensor, Tensor):
      raise TypeError(f"cannot assign {type(tensor)} object to buffer {name!r}")
    else:
      self._buffers[name] = tensor
      if persistent:
        self._non_persistent_buffers_set.discard(name)
      else:
        self._non_persistent_buffers_set.add(name)
      if tensor is not None:
        Naming.get_instance().register_module_child_attr(self, name, tensor)

  def apply(self: T, fn: Callable[['Module'], None]) -> T:
    with Naming.get_instance().push_module_apply(self):
      for module in self.children():
        module.apply(fn)
      fn(self)
    return self

  def register_forward_pre_hook(self, hook: Callable[..., None]) -> RemovableHandle:
    handle = RemovableHandle(self._forward_pre_hooks)
    self._forward_pre_hooks[handle.id] = hook
    return handle

  def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                            missing_keys, unexpected_keys, error_msgs):
    import torch
    persistent_buffers = {k: v for k, v in self._buffers.items() if k not in self._non_persistent_buffers_set}
    local_name_params = itertools.chain(self._parameters.items(), persistent_buffers.items())
    local_state = {k: v for k, v in local_name_params if v is not None}

    for name, param in local_state.items():
      key = prefix + name
      if key in state_dict:
        input_param = state_dict[key]
        if isinstance(input_param, torch.Tensor):  # we allow this here
          input_param = Tensor(*input_param.shape, numpy_array=input_param.detach().cpu().numpy())

        # Backward compatibility: loading 1-dim tensor from 0.3.* to version 0.4+
        if len(param.shape) == 0 and len(input_param.shape) == 1:
          input_param = input_param[0]

        if input_param.shape != param.shape:
          # local shape should match the one in checkpoint
          error_msgs.append('size mismatch for {}: copying a param with shape {} from checkpoint, '
                            'the shape in current model is {}.'
                            .format(key, input_param.shape, param.shape))
          continue

        try:
          with no_grad():
            param.copy_(input_param)
        except Exception as ex:
          error_msgs.append('While copying the parameter named "{}", '
                            'whose dimensions in the model are {} and '
                            'whose dimensions in the checkpoint are {}, '
                            'an exception occurred : {}.'
                            .format(key, param.size(), input_param.size(), ex.args))
      elif strict:
        missing_keys.append(key)

    if strict:
      for key in state_dict.keys():
        if key.startswith(prefix):
          input_name = key[len(prefix):]
          input_name = input_name.split('.', 1)[0]  # get the name of param/buffer/child
          if input_name not in self._modules and input_name not in local_state:
            unexpected_keys.append(key)

  def load_state_dict(self, state_dict: Union[Dict[str, Tensor], Dict[str, Tensor]], strict: bool = True):
    missing_keys = []
    unexpected_keys = []
    error_msgs = []

    # copy state_dict so _load_from_state_dict can modify it
    metadata = getattr(state_dict, '_metadata', None)
    state_dict = state_dict.copy()
    if metadata is not None:
      state_dict._metadata = metadata

    def load(module, prefix=''):
      local_metadata = {} if metadata is None else metadata.get(prefix[:-1], {})
      module._load_from_state_dict(
        state_dict, prefix, local_metadata, True, missing_keys, unexpected_keys, error_msgs)
      for name, child in module._modules.items():
        if child is not None:
          load(child, prefix + name + '.')

    load(self)
    load = None  # break load->load reference cycle

    if strict:
      if len(unexpected_keys) > 0:
        error_msgs.insert(
          0, 'Unexpected key(s) in state_dict: {}. '.format(
            ', '.join('"{}"'.format(k) for k in unexpected_keys)))
      if len(missing_keys) > 0:
        error_msgs.insert(
          0, 'Missing key(s) in state_dict: {}. '.format(
            ', '.join('"{}"'.format(k) for k in missing_keys)))

    if len(error_msgs) > 0:
      raise RuntimeError('Error(s) in loading state_dict for {}:\n\t{}'.format(
        self.__class__.__name__, "\n\t".join(error_msgs)))

  def named_modules(self, memo: Optional[Set[Module]] = None, prefix: str = ''):
    if memo is None:
      memo = set()
    if self not in memo:
      memo.add(self)
      yield prefix, self
      for name, module in self._modules.items():
        if module is None:
          continue
        submodule_prefix = prefix + ('.' if prefix else '') + name
        for m in module.named_modules(memo, submodule_prefix):
          yield m

  def _named_members(self, get_members_fn, prefix='', recurse=True):
    memo = set()
    modules = self.named_modules(prefix=prefix) if recurse else [(prefix, self)]
    for module_prefix, module in modules:
      members = get_members_fn(module)
      for k, v in members:
        if v is None or v in memo:
          continue
        memo.add(v)
        name = module_prefix + ('.' if module_prefix else '') + k
        yield name, v

  def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
    for name, param in self.named_parameters(recurse=recurse):
      yield param

  def named_parameters(self, prefix: str = '', recurse: bool = True) -> Iterator[Tuple[str, Tensor]]:
    gen = self._named_members(
      lambda module: module._parameters.items(),
      prefix=prefix, recurse=recurse)
    for elem in gen:
      yield elem

  def buffers(self, recurse: bool = True) -> Iterator[Tensor]:
    for name, buf in self.named_buffers(recurse=recurse):
      yield buf

  def named_buffers(self, prefix: str = '', recurse: bool = True) -> Iterator[Tuple[str, Tensor]]:
    gen = self._named_members(
      lambda module: module._buffers.items(),
      prefix=prefix, recurse=recurse)
    for elem in gen:
      yield elem

  def eval(self):
    return self  # ignore

  def train(self, arg):
    return self  # ignore

  def to(self, *args):
    return self  # ignore

  @property
  def training(self):
    return False

  def __call__(self, *input: Tensor, **kwargs):
    naming = Naming.get_instance()
    with naming.push_module_context(self):
      assert naming.wrap_to_returnn_enabled
      for hook in self._forward_pre_hooks.values():
        result = hook(self, input)
        if result is not None:
          if not isinstance(result, tuple):
            result = (result,)
          input = result
      with naming.push_module_call(module=self, inputs_args=input, inputs_kwargs=kwargs) as call_entry:
        res = call_entry.apply_call()
      return res

  def forward(self, *inputs: Tensor, **kwargs) -> Tensor:
    """
    Note:

    Normally we should never get here.

    However, there is one exception:
    We can still get here, in case some user code overrides some of the Torch modules, e.g. as in::

      class LayerNorm(torch.nn.LayerNorm):
          def __init__(self, nout, dim=-1):
              super(LayerNorm, self).__init__(nout, eps=1e-12)
              self.dim = dim
          def forward(self, x):
              if self.dim == -1:
                  return super(LayerNorm, self).forward(x)
              return super(LayerNorm, self).forward(x.transpose(1, -1)).transpose(1, -1)

    In that case, has_torch_forward() will return True (which is correct and expected),
    and then the forward() will be called, and that will lead to a call here.
    So we are expected to be in a subnetwork.
    But now we want to get the behavior of the original Torch module.
    """
    self_ = self
    orig_import_params_torch_to_returnn = self.import_params_torch_to_returnn

    def import_params_torch_to_returnn(*, layer: SubnetworkLayer, torch_module):
      assert isinstance(layer, SubnetworkLayer)
      sub_layer = layer.subnetwork.layers[self_.get_returnn_name()]
      orig_import_params_torch_to_returnn(layer=sub_layer, torch_module=torch_module)

    setattr(self, "import_params_torch_to_returnn", import_params_torch_to_returnn)
    setattr(self, "has_torch_forward", lambda: False)

    class _WrappedBaseClass(Module):
      is_original_torch_module = False

      def create_returnn_layer_dict(self, *inputs: Tensor, **kwargs) -> Dict[str, Any]:
        return self_.create_returnn_layer_dict(*inputs, **kwargs)

      def make_structured_returnn_output(self, output, *inputs_args, **inputs_kwargs):
        return self_.make_structured_returnn_output(output, *inputs_args, **inputs_kwargs)

      def get_returnn_name(self) -> str:
        return self_.get_returnn_name()

    _WrappedBaseClass.__qualname__ = f"_{self.__class__.__qualname__}_WrappedBaseClass"  # doesn't matter, but nicer
    _WrappedBaseClass.__name__ = f"_{self.__class__.__name__}_WrappedBaseClass"  # doesn't matter, but nicer
    wrapped_mod = _WrappedBaseClass()
    return wrapped_mod(*inputs)

  def create_returnn_layer_dict(self, *inputs: Tensor, **kwargs) -> Dict[str, Any]:
    raise Exception("should not get here")

  def make_structured_returnn_output(self, output: Tensor, *inputs: Tensor, **inputs_kwargs
                                     ) -> Union[Tensor, Tuple[Tensor], Any]:
    """
    This can be overridden alongside with `create_returnn_layer_dict`.
    In case the original `forward` would return not a single tensor but some tuple or dict or other nested structure,
    override this to match the original `forward` return structure.
    """
    return output

  @classmethod
  def has_torch_forward(cls) -> bool:
    """
    When we return True here, it implies that this module has a `forward` function,
    which would be wrapped via a RETURNN subnetwork.

    When we return False, it implies that this module has a `create_returnn_layer_dict` function,
    and would be wrapped directly via the corresponding RETURNN layer.
    """
    if cls.create_returnn_layer_dict is Module.create_returnn_layer_dict:
      return True  # always assume that the user module has custom forward code, even if not cls.forward
    return cls.forward is not Module.forward

  def check_returnn_layer(self, layer: LayerBase):
    """
    You can override this function to perform extra checks on the constructed RETURNN layer,
    e.g. the right input dimension.

    :param LayerBase layer: the constructed layer
    """

  def import_params_torch_to_returnn(self, *, layer: LayerBase, torch_module):
    """
    Override this function to import parameters from PyTorch.
    This only makes sense for modules available in PyTorch itself,
    with flag ``is_original_torch_module=True``.
    """

  @staticmethod
  def _check_call_returnn_input_to_prev_torch(call: CallEntry, tensor: TensorEntry, torch_values: numpy.ndarray):
    naming = Naming.get_instance()
    # Search backward to resolve all deps -- either to const, input, or already checked.
    # We keep all sizes separate, because we keep them in all cases,
    # as the RETURNN size optimization might mess up the dependency order.
    feed_dict = {}
    sizes_feed_dict = {}
    visited = set()
    queue = [tensor]
    while queue:
      queue_ = []
      for t in queue:
        if not isinstance(t, TensorEntry):
          continue
        if t in visited:
          continue
        visited.add(t)
        if t.validated_to_torch:
          pass
        elif t.is_const:
          print(f"**** validate: add const tensor {t}")
          t.validated_to_torch = True  # Skip check next time.
          # No need to feed, this should be const.
          t.validated_to_torch_tf_feed_dict = {}
          t.validated_to_torch_tf_sizes_feed_dict = {}
        elif t.is_input:
          print(f"**** validate: add network input tensor {t}")
          t.validated_to_torch = True
          t.validated_to_torch_tf_feed_dict = {}
          # Should be set via register_input.
          t_ = t.tensor()
          assert isinstance(t_, Tensor)
          data = t_.numpy()
          t.validated_to_torch_tf_feed_dict[t.returnn_data.placeholder] = data
          t.validated_to_torch_tf_sizes_feed_dict = {}
          for i, size in t.returnn_data.size_placeholder.items():
            batch_dim = data.shape[t.returnn_data.batch_dim_axis]
            size_ = data.shape[t.returnn_data.get_batch_axis(i)]
            t.validated_to_torch_tf_sizes_feed_dict[size] = numpy.array([size_] * batch_dim, dtype=numpy.int32)
          t.validated_to_torch_tf_feed_dict.update(t.validated_to_torch_tf_sizes_feed_dict)
        else:
          assert len(t.output_from_calls) >= 1
          call_ = t.output_from_calls[0]
          queue_ += call_.inputs_flat
          continue

        assert t.validated_to_torch
        feed_dict.update(t.validated_to_torch_tf_feed_dict)
        sizes_feed_dict.update(t.validated_to_torch_tf_sizes_feed_dict)

      queue = queue_

    print(f"**** validate: add call {call} input tensor {tensor}")
    session = tf.compat.v1.get_default_session()
    returnn_output_np_, output_sizes = session.run(
      (tensor.returnn_data.placeholder,
       tensor.returnn_data.size_placeholder.as_dict()),
      feed_dict=feed_dict)
    assert isinstance(returnn_output_np_, numpy.ndarray)
    tensor.validated_to_torch = True
    tensor.validated_to_torch_tf_feed_dict = {
      tensor.returnn_data.placeholder: returnn_output_np_}
    out_size_feed_dict = {
      tensor.returnn_data.size_placeholder[i]: output_sizes[i]
      for i in tensor.returnn_data.size_placeholder}
    out_size_feed_dict.update(sizes_feed_dict)
    tensor.validated_to_torch_tf_feed_dict.update(out_size_feed_dict)
    tensor.validated_to_torch_tf_sizes_feed_dict = out_size_feed_dict
    if any(call_.module.module._forward_feed_dict_deps for call_ in tensor.output_from_calls):
      tensor.validated_to_torch_tf_feed_dict.update(feed_dict)
    # Check now against Torch reference.
    torch_axis_from_returnn_axis = {i: j for (j, i) in tensor.returnn_axis_from_torch_axis.items()}
    assert len(torch_values.shape) == tensor.returnn_data.batch_ndim
    torch_values_ = torch_values.transpose(*[torch_axis_from_returnn_axis[i] for i in range(torch_values.ndim)])
    numpy.testing.assert_allclose(
      returnn_output_np_, torch_values_,
      err_msg=f"call {call} input tensor {tensor} check failed",
      **naming.validate_allclose_kwargs)

  @staticmethod
  def check_call_returnn_outputs_to_prev_torch(call: CallEntry, *, update_ops: List[tf.Operation]):
    naming = Naming.get_instance()
    assert naming.wrap_to_returnn_enabled and naming.import_params_from_torch_namespace and call.returnn_layer
    torch_naming = naming.import_params_from_torch_namespace
    mod_entry = call.module
    mod = mod_entry.module
    call_idx = naming.get_module_call_idx(module=mod, call=call)
    mod_abs_name = naming.get_module_abs_id_name(mod)
    # If the following throws an exception, maybe this module was marked with is_original_torch_module=True,
    # but actually it does not exist in Torch, or would not be used like this.
    # E.g. in the functional API, any created modules should use `as_returnn_torch_functional()`.
    torch_mod = torch_naming.get_module_by_abs_id_name(mod_abs_name)
    torch_mod_calls = torch_naming.get_module_calls(torch_mod)
    assert call_idx < len(torch_mod_calls)
    torch_mod_call = torch_mod_calls[call_idx]
    assert torch_mod_call.orig_outputs is not None and torch_mod_call.orig_inputs_flat is not None
    assert len(call.inputs_flat) == len(torch_mod_call.orig_inputs_flat)
    feed_dict = {}
    sizes_feed_dict = {}
    for tensor_entry, torch_input in zip(call.inputs_flat, torch_mod_call.orig_inputs_flat):
      if not isinstance(tensor_entry, TensorEntry):
        continue
      torch_input_np = torch_input.detach().cpu().numpy()
      Module._check_call_returnn_input_to_prev_torch(
        call=call, tensor=tensor_entry, torch_values=torch_input_np)
      feed_dict.update(tensor_entry.validated_to_torch_tf_feed_dict)
      sizes_feed_dict.update(tensor_entry.validated_to_torch_tf_sizes_feed_dict)
    session = tf.compat.v1.get_default_session()
    nest.assert_same_structure(call.outputs, torch_mod_call.orig_outputs)
    assert len(call.outputs_flat) == len(torch_mod_call.orig_outputs_flat)
    for i, x in enumerate(call.outputs_flat):
      idx_repr = f" {i + 1}/{len(call.outputs_flat)}" if len(call.outputs_flat) > 1 else ""
      print(f"**** validate: add call {call} output{idx_repr} tensor {x}")
    out, _ = session.run((
      [(x.returnn_data.placeholder, x.returnn_data.size_placeholder.as_dict()) if x is not None else () for
       x in call.outputs_flat], update_ops),
      feed_dict=feed_dict)
    for out_idx, returnn_output_tensor_entry in enumerate(call.outputs_flat):
      if returnn_output_tensor_entry is None:
        continue
      idx_repr = f" {out_idx + 1}/{len(call.outputs_flat)}" if len(call.outputs_flat) > 1 else ""
      returnn_output_np_, output_sizes = out[out_idx]
      assert isinstance(returnn_output_np_, numpy.ndarray)
      returnn_output_tensor_entry.validated_to_torch = True
      returnn_output_tensor_entry.validated_to_torch_tf_feed_dict = {
        returnn_output_tensor_entry.returnn_data.placeholder: returnn_output_np_}
      if call.module.module._forward_feed_dict_deps:
        returnn_output_tensor_entry.validated_to_torch_tf_feed_dict.update(feed_dict)
      out_size_feed_dict = {
        returnn_output_tensor_entry.returnn_data.size_placeholder[i]: output_sizes[i]
        for i in returnn_output_tensor_entry.returnn_data.size_placeholder}
      out_size_feed_dict.update(sizes_feed_dict)
      returnn_output_tensor_entry.validated_to_torch_tf_feed_dict.update(out_size_feed_dict)
      returnn_output_tensor_entry.validated_to_torch_tf_sizes_feed_dict = out_size_feed_dict
      returnn_output_np = returnn_output_np_.transpose(*[
        returnn_output_tensor_entry.returnn_axis_from_torch_axis[i] for i in range(returnn_output_np_.ndim)])
      torch_out_np = torch_mod_call.orig_outputs_flat[out_idx].detach().cpu().numpy()
      assert torch_out_np.shape == tuple(int(i) for i in returnn_output_tensor_entry.tensor().shape), (
        f"torch shape of TensorEntry does not match real torch shape: "
        f"{torch_out_np.shape} vs. {returnn_output_tensor_entry}")
      error_msg_info = [f"RETURNN layer: {call.returnn_layer}", f"Torch module: {torch_mod}"]
      if returnn_output_np.shape != torch_out_np.shape:
        error_msg_info += [
          f"ERROR: Output{idx_repr} shape mismatch",
          f"  RETURNN output{idx_repr} data: {call.outputs_flat[out_idx].returnn_data}",
          f"  RETURNN output{idx_repr} shape: {returnn_output_np_.shape},",
          f"  RETURNN output{idx_repr} shape (transposed to Torch): {returnn_output_np.shape},"
          f" (RETURNN<-Torch axis mapping {returnn_output_tensor_entry.returnn_axis_from_torch_axis})",
          f"  Torch output shape: {torch_out_np.shape}"]
        for i, (tensor_entry, torch_input) in enumerate(zip(call.inputs_flat, torch_mod_call.orig_inputs_flat)):
          error_msg_info += [f"input {i + 1}/{len(call.inputs_flat)}:"]
          if not isinstance(tensor_entry, TensorEntry):
            error_msg_info += ["  (non tensor)"]
            continue
          torch_axis_from_returnn_axis = {i: j for (j, i) in tensor_entry.returnn_axis_from_torch_axis.items()}
          torch_input_np_ = torch_input.detach().cpu().numpy()
          assert isinstance(torch_input_np_, numpy.ndarray)
          assert len(torch_input_np_.shape) == tensor_entry.returnn_data.batch_ndim
          torch_input_np = torch_input_np_.transpose(
            *[torch_axis_from_returnn_axis[i] for i in range(torch_input_np_.ndim)])
          error_msg_info += [
            f"  RETURNN input data: {tensor_entry.returnn_data}",
            f"  RETURNN input shape (Torch axis order): {torch_input_np_.shape}",
            f"  RETURNN input shape (transposed to RETURNN): {torch_input_np.shape}"
            f" (Torch<-RETURNN axis mapping {torch_axis_from_returnn_axis})"]
      else:
        is_close_arr = numpy.isclose(returnn_output_np, torch_out_np, **naming.validate_allclose_kwargs)
        if not numpy.all(is_close_arr):
          idx = numpy.argmax(numpy.abs(returnn_output_np - torch_out_np))
          idx_ = numpy.unravel_index(idx, shape=returnn_output_np.shape)
          error_msg_info += [
            f"  RETURNN output{idx_repr} min/max: {numpy.min(returnn_output_np), numpy.max(returnn_output_np)}",
            f"  Torch output{idx_repr} min/max: {numpy.min(torch_out_np), numpy.max(torch_out_np)}",
            f"  Biggest mismatch at idx {idx_}, RETURNN {returnn_output_np[idx_]} vs Torch {torch_out_np[idx_]}",
          ]
      numpy.testing.assert_allclose(
        returnn_output_np, torch_out_np,
        err_msg="\n".join(error_msg_info),
        **naming.validate_allclose_kwargs)

  @staticmethod
  def _get_input_layer_name(input: Tensor) -> str:
    naming = Naming.get_instance()
    assert naming.module_call_stack
    top_call_entry = naming.module_call_stack[-1]
    parent_namespace = top_call_entry.namespace.parent
    x = naming.get_tensor(input)
    naming.prepare_tensor_as_input(x, parent_namespace=parent_namespace)
    # Note: If name_for_tensor fails, it means the tensor was not registered properly.
    name, _ = parent_namespace.name_for_tensor(x)
    return name

  @staticmethod
  def _assert_spatial_axes_in_order(input: Tensor):
    entry = Naming.get_instance().register_tensor(input)
    assert entry.returnn_data
    spatial_axes = entry.returnn_data.get_spatial_batch_axes()
    returnn_axis_to_torch_axis = {i: j for (j, i) in entry.returnn_axis_from_torch_axis.items()}
    spatial_axes_torch = [returnn_axis_to_torch_axis[i] for i in spatial_axes]
    assert sorted(spatial_axes_torch) == spatial_axes_torch

  @staticmethod
  def _assert_axes_in_order(input: Tensor, *, dims: Collection[int]):
    entry = Naming.get_instance().register_tensor(input)
    assert entry.returnn_data
    spatial_axes_returnn = [entry.returnn_axis_from_torch_axis[i] for i in sorted(dims)]
    assert sorted(spatial_axes_returnn) == spatial_axes_returnn

  @staticmethod
  def _get_input_axis_to_returnn(input: Tensor, axis: int) -> str:
    return Naming.get_instance().register_tensor(input).get_returnn_axis_description(axis)

  def make_output_tensor_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase) -> Tensor:
    naming = Naming.get_instance()
    torch_shape, returnn_axis_from_torch_axis = self._get_output_shape_from_returnn(
      inputs_flat=inputs_flat, layer=layer)
    for i in range(len(torch_shape)):
      assert layer.output.batch_shape[returnn_axis_from_torch_axis[i]] in {None, torch_shape[i]}
    is_const = False
    numpy_array = None
    inputs_entries = [
      naming.tensors[x] if isinstance(x, Tensor) else None for x in inputs_flat]  # type: List[Optional[TensorEntry]]
    if all([x.is_const if x else True for x in inputs_entries]):
      # Only have const input.
      # Evaluate layer, because this const might be used in certain operation (e.g. predefined filter for conv).
      session = tf.compat.v1.get_default_session()
      feed_dict = {}
      for x in inputs_entries:
        if not x:
          continue
        value = x.tensor().numpy()
        value = numpy.transpose(value, [x.torch_axis_from_returnn_axis[i] for i in range(value.ndim)])
        feed_dict[x.returnn_data.placeholder] = value
      numpy_array = session.run(layer.output.placeholder, feed_dict=feed_dict)
      numpy_array = numpy.transpose(numpy_array, [returnn_axis_from_torch_axis[i] for i in range(numpy_array.ndim)])
      is_const = True
    tensor = Tensor(*torch_shape, numpy_array=numpy_array, dtype=layer.output.dtype)
    tensor_entry = naming.register_tensor(tensor)
    tensor_entry.is_const = is_const
    tensor_entry.returnn_data = layer.output
    tensor_entry.returnn_axis_from_torch_axis = returnn_axis_from_torch_axis
    return tensor

  @staticmethod
  def _base_get_output_shape_from_returnn(inputs_flat: List[Tensor], layer: LayerBase
                                          ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    :return: (torch_shape, returnn_axis_from_torch_axis).
      Torch shape how it would have looked when this would be processed within Torch.
      The RETURNN layer.output shape (order of axes) might look different.

    This is a bit tricky.
    If axes got reordered (RETURNN does that for efficiency, and then returns as-is),
    Torch code would not expect this.
    We need a mapping of RETURNN axis -> Torch axis.
    We can automatically infer this, but this is a bit involved.
    """
    # We could also use functions like Data.get_batch_shape_dim_tags, Data.get_common_data,
    # DimensionTag.get_all_dimension_tags, etc.
    # However, this is simpler, and also more consistent with get_returnn_axis_description.

    def _get_spatial_dim_tag_and_single_index(data: Data, axis: int) -> Tuple[DimensionTag, int]:
      """
      :return: (RETURNN dim tag, spatial dim tag index (for stag-single))
      """
      dim_tag_ = data.get_dim_tag(axis)
      idx_ = data.get_axes_by_tag_name(dim_tag_.description).index(axis)
      return dim_tag_, idx_

    out_returnn_axis_to_torch_axis = {}
    # Torch would maybe have operated on [B,D_in,T_in] input, and produce [B,D_out,T_out] output.
    naming = Naming.get_instance()
    batch_size = None
    # dim_tag_ext -> in spatial idx, Torch dim where dim_tag_ext is from _get_spatial_dim_tag_and_single_index
    dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim = OrderedDict()
    for input in inputs_flat:
      if not isinstance(input, Tensor):
        continue
      if not input.shape:
        continue  # skip scalars
      x = naming.tensors[input]
      assert isinstance(x, TensorEntry)
      assert x.returnn_data and x.returnn_axis_from_torch_axis is not None
      if x.returnn_data.have_batch_axis():
        batch_size = input.shape[x.torch_axis_from_returnn_axis[x.returnn_data.batch_dim_axis]]
      for i in x.returnn_data.get_dynamic_axes():
        dim_tag_ext = _get_spatial_dim_tag_and_single_index(x.returnn_data, i)
        assert i in x.returnn_data.get_spatial_batch_axes()
        spatial_idx = x.returnn_data.get_spatial_batch_axes().index(i)
        torch_dim = input.shape[x.torch_axis_from_returnn_axis[i]]
        if dim_tag_ext not in dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim:
          dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim[dim_tag_ext] = (spatial_idx, torch_dim)

      # Find mapping to layer_output_shape_meta.
      mapping_out_to_in = {}
      for out_axis in range(layer.output.batch_ndim):
        if out_axis == layer.output.batch_dim_axis:
          if x.returnn_data.have_batch_axis():
            mapping_out_to_in[out_axis] = x.returnn_data.batch_dim_axis
          else:
            mapping_out_to_in[out_axis] = None  # new axis
          continue
        if out_axis == layer.output.feature_dim_axis:
          if x.returnn_data.have_feature_axis():
            mapping_out_to_in[out_axis] = x.returnn_data.feature_dim_axis
          else:
            mapping_out_to_in[out_axis] = None  # new axis
          continue
        if out_axis in layer.output.get_dynamic_axes():
          dim_tag_ext = _get_spatial_dim_tag_and_single_index(layer.output, out_axis)
          if dim_tag_ext in dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim:
            in_spatial_idx, _ = dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim[dim_tag_ext]
            mapping_out_to_in[out_axis] = x.returnn_data.get_spatial_batch_axes()[in_spatial_idx]
            continue
        assert out_axis in layer.output.get_spatial_batch_axes()
        out_spatial_idx = layer.output.get_spatial_batch_axes().index(out_axis)
        if len(x.returnn_data.get_spatial_batch_axes()) == len(layer.output.get_spatial_batch_axes()):
          mapping_out_to_in[out_axis] = x.returnn_data.get_spatial_batch_axes()[out_spatial_idx]
          continue
        # Just skip other cases now.

      in_values = [j for (i, j) in sorted(mapping_out_to_in.items()) if j is not None]
      if in_values != sorted(in_values):  # do we have some reordering?
        rem_out = [i for i in range(layer.output.batch_ndim) if i not in mapping_out_to_in]
        rem_in = [i for i in range(x.returnn_data.batch_ndim) if i not in in_values]
        assert len(rem_out) == len(rem_in)  # assumption, otherwise no idea what to do
        assert None not in list(mapping_out_to_in.values())  # no new axis
        mapping_out_to_in.update({i: j for (i, j) in zip(rem_out, rem_in)})

        # Assume no reordering happened on the Torch site.
        for returnn_out_axis, returnn_in_axis in mapping_out_to_in.items():
          if returnn_in_axis is not None:
            out_returnn_axis_to_torch_axis[returnn_out_axis] = x.torch_axis_from_returnn_axis[returnn_in_axis]

      else:  # same order, but maybe some dims added or removed
        if layer.output.batch_ndim == x.returnn_data.batch_ndim:  # no dim added/removed
          for returnn_out_axis, returnn_in_axis in mapping_out_to_in.items():
            if returnn_in_axis is not None:
              out_returnn_axis_to_torch_axis[returnn_out_axis] = x.torch_axis_from_returnn_axis[returnn_in_axis]
        elif layer.output.batch_ndim < x.returnn_data.batch_ndim:  # dim removed
          assert all(v is not None for k, v in mapping_out_to_in.items())  # None implies new dim, which is wrong
          if set(mapping_out_to_in.keys()) == set(range(layer.output.batch_ndim)):  # all out dims mapped to input
            covered_in_returnn_axes = set(mapping_out_to_in.values())
            removed_in_returnn_axes = [i for i in range(x.returnn_data.batch_ndim) if i not in covered_in_returnn_axes]
            removed_in_torch_axes = set(x.torch_axis_from_returnn_axis[i] for i in removed_in_returnn_axes)
            for returnn_out_axis, returnn_in_axis in mapping_out_to_in.items():
              if returnn_in_axis is not None:
                in_torch_axis = x.torch_axis_from_returnn_axis[returnn_in_axis]
                assert in_torch_axis not in removed_in_torch_axes
                in_torch_axis -= len([i for i in range(in_torch_axis) if i in removed_in_torch_axes])
                out_returnn_axis_to_torch_axis[returnn_out_axis] = in_torch_axis

      break
    assert all(0 <= d < layer.output.batch_ndim for d in out_returnn_axis_to_torch_axis.values())
    assert len(set(out_returnn_axis_to_torch_axis.values())) == len(out_returnn_axis_to_torch_axis)
    # The remaining axes should be the spatial axes.
    rem_torch_axes = set(range(layer.output.batch_ndim)).difference(set(out_returnn_axis_to_torch_axis.values()))
    rem_returnn_axes = set(range(layer.output.batch_ndim)).difference(set(out_returnn_axis_to_torch_axis.keys()))
    assert len(rem_torch_axes) == len(rem_returnn_axes)
    rem_torch_axes_ = sorted(rem_torch_axes)
    rem_returnn_axes_ = sorted(rem_returnn_axes)

    out_shape = list(layer.output.batch_shape)
    if layer.output.have_batch_axis():
      assert batch_size is not None
      out_shape[layer.output.batch_dim_axis] = batch_size
    if layer.output.get_dynamic_axes():
      assert dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim
      for i in layer.output.get_dynamic_axes():
        dim_tag_ext = _get_spatial_dim_tag_and_single_index(layer.output, i)
        if dim_tag_ext in dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim:
          in_spatial_idx, out_shape[i] = dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim[dim_tag_ext]
          if i in rem_returnn_axes:
            out_spatial_idx = rem_returnn_axes_.index(i)
            if in_spatial_idx != out_spatial_idx:
              out_returnn_axis_to_torch_axis[i] = rem_torch_axes_[in_spatial_idx]
              rem_returnn_axes.remove(i)
              rem_torch_axes.remove(rem_torch_axes_[in_spatial_idx])
        else:
          # Assume same order.
          assert len(layer.output.get_dynamic_axes()) == len(dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim)
          out_shape[i] = (
            list(dyn_size_dim_tag_ext_to_spatial_idx_and_torch_dim.values())[layer.output.get_dynamic_axes().index(i)][1])
    assert all(d for d in out_shape)

    for i, j in zip(sorted(rem_returnn_axes), sorted(rem_torch_axes)):
      assert i not in out_returnn_axis_to_torch_axis
      out_returnn_axis_to_torch_axis[i] = j
    assert (
        len(set(out_returnn_axis_to_torch_axis.values())) ==
        len(set(out_returnn_axis_to_torch_axis.keys())) ==
        layer.output.batch_ndim)
    torch_axis_to_returnn = {i: j for (j, i) in out_returnn_axis_to_torch_axis.items()}
    assert len(torch_axis_to_returnn) == layer.output.batch_ndim
    out_shape = [out_shape[torch_axis_to_returnn[i]] for i in range(layer.output.batch_ndim)]
    out_returnn_axis_from_torch_axis = torch_axis_to_returnn
    return tuple(out_shape), out_returnn_axis_from_torch_axis

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    :return: (torch_shape, returnn_axis_from_torch_axis).
      Torch shape how it would have looked when this would be processed within Torch.
      The RETURNN layer.output shape (order of axes) might look different.

    Note about returnn_axis_from_torch_axis:
    This is when you would use `tf.transpose` or `numpy.transpose`,
    and you can think of `returnn_axis_from_torch_axis` as the `perm` argument, as a list.
    The dict is like {torch_axis: returnn_axis}, i.e. Torch axis -> RETURNN axis.
    You would use this if you want to transpose RETURNN axes **to** Torch axes
    (i.e. the other way around).

    This is a bit tricky.
    If axes got reordered (RETURNN does that for efficiency, and then returns as-is),
    Torch code would not expect this.
    We need a mapping of RETURNN axis -> Torch axis.
    We can automatically infer this, but this is a bit involved.
    """
    return self._base_get_output_shape_from_returnn(inputs_flat=inputs_flat, layer=layer)

  def as_returnn_torch_functional(self):
    """
    Call this to mark this module such that this would not be used when used with the original Torch.
    Usually this is via the functional API, which calls back to standard Torch modules.

    :return: self
    """
    self.is_original_torch_module = False
    return self

  @classmethod
  def _make_returnn_dummy_input(cls, data: Data) -> numpy.ndarray:
    if getattr(cls, "returnn_dummy_input_shape", None) is not None:
      shape = cls.returnn_dummy_input_shape
    else:
      some_primes = (3, 5, 7, 11, 13)  # use primes for dynamic dims, just nicer to identify in logs
      dynamic_axes = [i for i, dim in enumerate(data.batch_shape) if dim is None]
      assert len(dynamic_axes) <= len(some_primes)  # just not implemented otherwise
      shape = list(data.batch_shape)
      for i, j in enumerate(dynamic_axes):
        shape[j] = some_primes[i]
    return numpy.zeros(shape, dtype=data.dtype)

  def _returnn_dummy_call(self, *returnn_inputs: Dict[str, Any]) -> Naming:
    from pytorch_to_returnn.torch import from_numpy
    naming = Naming.get_instance()
    returnn_datas = []
    for i, kwargs in enumerate(returnn_inputs):
      kwargs = kwargs.copy()
      if "name" not in kwargs:
        kwargs["name"] = "data" if i == 0 else f"data:{i}"
      x = Data(**kwargs)
      returnn_datas.append(x)
    dummy_inputs_np = [self._make_returnn_dummy_input(x) for x in returnn_datas]
    dummy_inputs_torch = [from_numpy(x) for x in dummy_inputs_np]
    for i in range(len(returnn_inputs)):
      naming.register_input(tensor=dummy_inputs_torch[i], returnn_data=returnn_datas[i])
    out = self(*dummy_inputs_torch)
    assert isinstance(out, Tensor)
    naming.register_output(out)
    return naming

  def as_returnn_net_dict(self, *returnn_inputs: Dict[str, Any]):
    return self._returnn_dummy_call(*returnn_inputs).root_namespace.dump_as_returnn_net_dict()

  def as_returnn_layer_dict(self, *returnn_inputs: Dict[str, Any]):
    return self._returnn_dummy_call(*returnn_inputs).root_namespace.dump_as_returnn_layer_dict()




from ...tensor import Tensor
from .module import Module
from returnn.tf.layers.basic import LayerBase
from typing import Union, Tuple, Optional, List, Dict
from .utils import _pair, _quadruple, _ntuple
from ..common_types import _size_2_t, _size_4_t, _size_6_t
from ....naming import Naming


class GenericPadNd(Module):
  padding: Tuple[int, ...] = None  # set by instance
  mode: str = None  # set by subclass
  value: float = None  # set by subclass instance
  nd: int = None  # set by subclass

  def __init__(self, *, padding: Tuple[int, ...], mode: Optional[str] = None, value: Optional[float] = None):
    super(GenericPadNd, self).__init__()
    self.padding = padding
    if mode is not None:
      self.mode = mode
    if value is not None:
      self.value = value

  def create_returnn_layer_dict(self, input: Tensor):
    assert self.mode
    assert self.mode != "replicate"  # not implemented
    assert len(self.padding) % 2 == 0, 'Padding needs to be a multiple of 2'

    naming = Naming.get_instance()
    input_naming = naming.tensors[input]

    padding_returnn = [(self.padding[2 * i], self.padding[2 * i + 1]) for i in range(len(self.padding) // 2)]
    axes_returnn = [
      input_naming.returnn_axis_from_torch_axis[i + input.ndim] for i in range(-1, -(len(self.padding) // 2) - 1, -1)]

    # remove axes with padding (0, 0) to simplify RETURNN network
    non_zero_padding_axes = [ax for ax in range(len(padding_returnn)) if padding_returnn[ax] != (0, 0)]
    axes_returnn = [axes_returnn[ax] for ax in non_zero_padding_axes]
    padding_returnn = [padding_returnn[ax] for ax in non_zero_padding_axes]

    if len(non_zero_padding_axes) == 0:
      return {"class": "copy", "from": self._get_input_layer_name(input)}

    d = {
      "class": "pad", "mode": self.mode,
      "axes": [self._get_input_axis_to_returnn(input, a) for a in axes_returnn],
      "padding": padding_returnn,
      "from": self._get_input_layer_name(input)}
    if self.mode == "constant":
      d["value"] = self.value
    return d

  def _get_output_shape_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    The size of the dynamic axes might be changed, so we have to take care of this here for the torch shape.
    """
    torch_shape, returnn_axis_from_torch_axis = super(GenericPadNd, self)._get_output_shape_from_returnn(
      inputs_flat=inputs_flat, layer=layer)
    assert len(inputs_flat) == 1
    torch_shape = list(inputs_flat[0].shape)
    for idx in range(len(self.padding) // 2):
      torch_shape[-1 - idx] += self.padding[2 * idx] + self.padding[2 * idx + 1]
    return tuple(torch_shape), returnn_axis_from_torch_axis


class _ConstantPadNd(GenericPadNd):
  mode = "constant"

  def __init__(self, *, padding: Tuple[int, ...], value: float) -> None:
    super(_ConstantPadNd, self).__init__(padding=padding)
    self.value = value


class ConstantPad1d(_ConstantPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t, value: float):
    super(ConstantPad1d, self).__init__(padding=_pair(padding), value=value)


class _ReflectionPadNd(GenericPadNd):
  mode = "reflect"


class ReflectionPad1d(_ReflectionPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t):
    super(ReflectionPad1d, self).__init__(padding=_pair(padding))


class _ReplicationPadNd(GenericPadNd):
  mode = "replicate"


class ReplicationPad1d(_ReplicationPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t) -> None:
    super(ReplicationPad1d, self).__init__(padding=_pair(padding))


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from .tensor import *
from ._C import *
from .nn.functional import *
from .serialization import load
from .autograd import *
from . import nn
from . import cuda
from . import onnx
from . import jit

__version__ = "1.8.1"



#!/usr/bin/env python3

"""
Copied many things from https://github.com/pytorch/examples/blob/master/mnist/main.py.
"""

import _setup_env  # noqa
import better_exchook
import typing
from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn
import torch
import torch.utils.data
from torchvision import datasets, transforms
import argparse
import os


def model_func(wrapped_import, inputs):
  if wrapped_import:
    model = wrapped_import("model")
  else:
    import model
  net = model.Net()
  net = net.eval()  # disable dropout
  return net(inputs)


my_dir = os.path.dirname(os.path.abspath(__file__))


def main():
  # Training settings
  parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
  parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                      help='input batch size for training (default: 64)')
  parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                      help='input batch size for testing (default: 1000)')
  parser.add_argument('--epochs', type=int, default=14, metavar='N',
                      help='number of epochs to train (default: 14)')
  parser.add_argument('--lr', type=float, default=1.0, metavar='LR',
                      help='learning rate (default: 1.0)')
  parser.add_argument('--gamma', type=float, default=0.7, metavar='M',
                      help='Learning rate step gamma (default: 0.7)')
  parser.add_argument('--no-cuda', action='store_true', default=False,
                      help='disables CUDA training')
  parser.add_argument('--dry-run', action='store_true', default=False,
                      help='quickly check a single pass')
  parser.add_argument('--seed', type=int, default=1, metavar='S',
                      help='random seed (default: 1)')
  parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                      help='how many batches to wait before logging training status')
  parser.add_argument('--save-model', action='store_true', default=False,
                      help='For Saving the current Model')
  args = parser.parse_args()
  use_cuda = not args.no_cuda and torch.cuda.is_available()

  torch.manual_seed(args.seed)

  device = torch.device("cuda" if use_cuda else "cpu")

  train_kwargs = {'batch_size': args.batch_size}
  test_kwargs = {'batch_size': args.test_batch_size}
  if use_cuda:
    cuda_kwargs = {'num_workers': 1, 'pin_memory': True, 'shuffle': True}
    train_kwargs.update(cuda_kwargs)
    test_kwargs.update(cuda_kwargs)

  transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.1307,), (0.3081,))
  ])
  dataset1 = datasets.MNIST(f'{my_dir}/../data', train=True, download=True, transform=transform)
  dataset2 = datasets.MNIST(f'{my_dir}/../data', train=False, transform=transform)
  train_loader = torch.utils.data.DataLoader(dataset1, **train_kwargs)
  test_loader = torch.utils.data.DataLoader(dataset2, **test_kwargs)

  data, target = next(iter(train_loader))
  data_np = data.numpy()

  verify_torch_and_convert_to_returnn(
    model_func, inputs=data_np, inputs_data_kwargs={"shape": (1, 28, 28)})


if __name__ == '__main__':
  better_exchook.install()
  main()



def export(fn):
  """Dummy function to avoid import errors"""
  pass


def unused(fn):
  """Dummy function to avoid import errors"""
  pass


def is_scripting() -> bool:
  return False  # stub


def _unwrap_optional(x):
    assert x is not None, "Unwrapping null optional"
    return x




# coding: utf-8
import argparse
import time
import math
import os
import torch
import torch.nn as nn
import torch.onnx

import data
import model

parser = argparse.ArgumentParser(description='PyTorch Wikitext-2 RNN/LSTM/GRU/Transformer Language Model')
parser.add_argument('--data', type=str, default='./data/wikitext-2',
                    help='location of the data corpus')
parser.add_argument('--model', type=str, default='LSTM',
                    help='type of recurrent net (RNN_TANH, RNN_RELU, LSTM, GRU, Transformer)')
parser.add_argument('--emsize', type=int, default=200,
                    help='size of word embeddings')
parser.add_argument('--nhid', type=int, default=200,
                    help='number of hidden units per layer')
parser.add_argument('--nlayers', type=int, default=2,
                    help='number of layers')
parser.add_argument('--lr', type=float, default=20,
                    help='initial learning rate')
parser.add_argument('--clip', type=float, default=0.25,
                    help='gradient clipping')
parser.add_argument('--epochs', type=int, default=40,
                    help='upper epoch limit')
parser.add_argument('--batch_size', type=int, default=20, metavar='N',
                    help='batch size')
parser.add_argument('--bptt', type=int, default=35,
                    help='sequence length')
parser.add_argument('--dropout', type=float, default=0.2,
                    help='dropout applied to layers (0 = no dropout)')
parser.add_argument('--tied', action='store_true',
                    help='tie the word embedding and softmax weights')
parser.add_argument('--seed', type=int, default=1111,
                    help='random seed')
parser.add_argument('--cuda', action='store_true',
                    help='use CUDA')
parser.add_argument('--log-interval', type=int, default=200, metavar='N',
                    help='report interval')
parser.add_argument('--save', type=str, default='model.pt',
                    help='path to save the final model')
parser.add_argument('--onnx-export', type=str, default='',
                    help='path to export the final model in onnx format')

parser.add_argument('--nhead', type=int, default=2,
                    help='the number of heads in the encoder/decoder of the transformer model')
parser.add_argument('--dry-run', action='store_true',
                    help='verify the code and the model')

args = parser.parse_args()

# Set the random seed manually for reproducibility.
torch.manual_seed(args.seed)
if torch.cuda.is_available():
    if not args.cuda:
        print("WARNING: You have a CUDA device, so you should probably run with --cuda")

device = torch.device("cuda" if args.cuda else "cpu")

###############################################################################
# Load data
###############################################################################

corpus = data.Corpus(args.data)

# Starting from sequential data, batchify arranges the dataset into columns.
# For instance, with the alphabet as the sequence and batch size 4, we'd get
# ┌ a g m s ┐
# │ b h n t │
# │ c i o u │
# │ d j p v │
# │ e k q w │
# └ f l r x ┘.
# These columns are treated as independent by the model, which means that the
# dependence of e. g. 'g' on 'f' can not be learned, but allows more efficient
# batch processing.

def batchify(data, bsz):
    # Work out how cleanly we can divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    return data.to(device)

eval_batch_size = 10
train_data = batchify(corpus.train, args.batch_size)
val_data = batchify(corpus.valid, eval_batch_size)
test_data = batchify(corpus.test, eval_batch_size)

###############################################################################
# Build the model
###############################################################################

ntokens = len(corpus.dictionary)
if args.model == 'Transformer':
    model = model.TransformerModel(ntokens, args.emsize, args.nhead, args.nhid, args.nlayers, args.dropout).to(device)
else:
    model = model.RNNModel(args.model, ntokens, args.emsize, args.nhid, args.nlayers, args.dropout, args.tied).to(device)

criterion = nn.NLLLoss()

###############################################################################
# Training code
###############################################################################

def repackage_hidden(h):
    """Wraps hidden states in new Tensors, to detach them from their history."""

    if isinstance(h, torch.Tensor):
        return h.detach()
    else:
        return tuple(repackage_hidden(v) for v in h)


# get_batch subdivides the source data into chunks of length args.bptt.
# If source is equal to the example output of the batchify function, with
# a bptt-limit of 2, we'd get the following two Variables for i = 0:
# ┌ a g m s ┐ ┌ b h n t ┐
# └ b h n t ┘ └ c i o u ┘
# Note that despite the name of the function, the subdivison of data is not
# done along the batch dimension (i.e. dimension 1), since that was handled
# by the batchify function. The chunks are along dimension 0, corresponding
# to the seq_len dimension in the LSTM.

def get_batch(source, i):
    seq_len = min(args.bptt, len(source) - 1 - i)
    data = source[i:i+seq_len]
    target = source[i+1:i+1+seq_len].view(-1)
    return data, target


def evaluate(data_source):
    # Turn on evaluation mode which disables dropout.
    model.eval()
    total_loss = 0.
    ntokens = len(corpus.dictionary)
    if args.model != 'Transformer':
        hidden = model.init_hidden(eval_batch_size)
    with torch.no_grad():
        for i in range(0, data_source.size(0) - 1, args.bptt):
            data, targets = get_batch(data_source, i)
            if args.model == 'Transformer':
                output = model(data)
                output = output.view(-1, ntokens)
            else:
                output, hidden = model(data, hidden)
                hidden = repackage_hidden(hidden)
            total_loss += len(data) * criterion(output, targets).item()
    return total_loss / (len(data_source) - 1)


def train():
    # Turn on training mode which enables dropout.
    model.train()
    total_loss = 0.
    start_time = time.time()
    ntokens = len(corpus.dictionary)
    if args.model != 'Transformer':
        hidden = model.init_hidden(args.batch_size)
    for batch, i in enumerate(range(0, train_data.size(0) - 1, args.bptt)):
        data, targets = get_batch(train_data, i)
        # Starting each batch, we detach the hidden state from how it was previously produced.
        # If we didn't, the model would try backpropagating all the way to start of the dataset.
        model.zero_grad()
        if args.model == 'Transformer':
            output = model(data)
            output = output.view(-1, ntokens)
        else:
            hidden = repackage_hidden(hidden)
            output, hidden = model(data, hidden)
        loss = criterion(output, targets)
        loss.backward()

        # `clip_grad_norm` helps prevent the exploding gradient problem in RNNs / LSTMs.
        torch.nn.utils.clip_grad_norm_(model.parameters(), args.clip)
        for p in model.parameters():
            p.data.add_(p.grad, alpha=-lr)

        total_loss += loss.item()

        if batch % args.log_interval == 0 and batch > 0:
            cur_loss = total_loss / args.log_interval
            elapsed = time.time() - start_time
            print('| epoch {:3d} | {:5d}/{:5d} batches | lr {:02.2f} | ms/batch {:5.2f} | '
                    'loss {:5.2f} | ppl {:8.2f}'.format(
                epoch, batch, len(train_data) // args.bptt, lr,
                elapsed * 1000 / args.log_interval, cur_loss, math.exp(cur_loss)))
            total_loss = 0
            start_time = time.time()
        if args.dry_run:
            break


def export_onnx(path, batch_size, seq_len):
    print('The model is also exported in ONNX format at {}'.
          format(os.path.realpath(args.onnx_export)))
    model.eval()
    dummy_input = torch.LongTensor(seq_len * batch_size).zero_().view(-1, batch_size).to(device)
    hidden = model.init_hidden(batch_size)
    torch.onnx.export(model, (dummy_input, hidden), path)


# Loop over epochs.
lr = args.lr
best_val_loss = None

# At any point you can hit Ctrl + C to break out of training early.
try:
    for epoch in range(1, args.epochs+1):
        epoch_start_time = time.time()
        train()
        val_loss = evaluate(val_data)
        print('-' * 89)
        print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | '
                'valid ppl {:8.2f}'.format(epoch, (time.time() - epoch_start_time),
                                           val_loss, math.exp(val_loss)))
        print('-' * 89)
        # Save the model if the validation loss is the best we've seen so far.
        if not best_val_loss or val_loss < best_val_loss:
            with open(args.save, 'wb') as f:
                torch.save(model, f)
            best_val_loss = val_loss
        else:
            # Anneal the learning rate if no improvement has been seen in the validation dataset.
            lr /= 4.0
except KeyboardInterrupt:
    print('-' * 89)
    print('Exiting from training early')

# Load the best saved model.
with open(args.save, 'rb') as f:
    model = torch.load(f)
    # after load the rnn params are not a continuous chunk of memory
    # this makes them a continuous chunk, and will speed up forward pass
    # Currently, only rnn model supports flatten_parameters function.
    if args.model in ['RNN_TANH', 'RNN_RELU', 'LSTM', 'GRU']:
        model.rnn.flatten_parameters()

# Run on test data.
test_loss = evaluate(test_data)
print('=' * 89)
print('| End of training | test loss {:5.2f} | test ppl {:8.2f}'.format(
    test_loss, math.exp(test_loss)))
print('=' * 89)

if len(args.onnx_export) > 0:
    # Export the model in ONNX format.
    export_onnx(args.onnx_export, batch_size=1, seq_len=args.bptt)




import ast
import linecache
from .. import log
from .mod_map import ModMap
from typing import Optional


class AstImportTransformer(ast.NodeTransformer):
  def __init__(self, mod_map: ModMap, src_filename: str):
    """
    :param base_mod_map: e.g.
      {"parallel_wavegan": "pytorch_to_returnn._wrapped_mods.parallel_wavegan",
       "torch": "pytorch_to_returnn._wrapped_mods.torch"}
      or
      {"parallel_wavegan": "pytorch_to_returnn._demo_mods.parallel_wavegan",
       "torch": "pytorch_to_returnn.torch"}
    :param src_filename: just for logging
    """
    self.mod_map = mod_map
    self.src_filename = src_filename

  def visit_Import(self, node: ast.Import):
    # https://docs.python.org/3/library/ast.html#ast.Import
    if not any([self.mod_map.should_wrap_mod_name(alias.name) for alias in node.names]):
      return node
    if len(node.names) >= 2:
      # Break down into multiple imports.
      res = []
      for alias in node.names:
        assert isinstance(alias, ast.alias)
        sub_node = ast.Import(names=[alias])
        ast.copy_location(sub_node, node)
        sub_node = self.visit_Import(sub_node)
        assert isinstance(sub_node, ast.Import)
        res.append(sub_node)
      return res
    assert len(node.names) == 1
    alias, = node.names
    assert isinstance(alias, ast.alias)
    if not self.mod_map.should_wrap_mod_name(alias.name):
      return node
    if log.Verbosity >= 10:
      log.unique_print("*** AST transform %r (%s)" % (_ast_get_source_segment(self.src_filename, node), ast.dump(node)))
    # E.g. `import parallel_wavegan.layers`
    # ->
    # `import pytorch_to_returnn._wrapped_mods.parallel_wavegan.layers`
    # If "." in mod name, also:
    # `from pytorch_to_returnn._wrapped_mods import parallel_wavegan`
    # If `import parallel_wavegan.layers as x`, simpler:
    if alias.asname:  # import torch as x -> import pytorch_to_returnn._wrapped_mods.torch as x
      new_node = ast.Import(
        names=[ast.alias(
          name=self.mod_map.map_mod_name(alias.name), asname=alias.asname)])
      ast.copy_location(new_node, node)
      return new_node
    # If "." in mod name, simpler:
    if "." not in alias.name:  # import torch -> import pytorch_to_returnn._wrapped_mods.torch as torch
      new_node = ast.Import(
        names=[ast.alias(
          name=self.mod_map.map_mod_name(alias.name), asname=alias.name)])
      ast.copy_location(new_node, node)
      return new_node
    assert not alias.asname
    new_import_node = ast.Import(names=[ast.alias(name=self.mod_map.map_mod_name(alias.name))])
    ast.copy_location(new_import_node, node)
    assert "." in alias.name
    base_mod_name = self.mod_map.find_base_mod_prefix(alias.name)  # e.g. "torch"
    assert "." not in base_mod_name  # just not implemented yet
    new_alias_node = ast.ImportFrom(  # from pytorch_to_returnn._wrapped_mods import torch
      module=self.mod_map.base_mod_map[base_mod_name].rpartition(".")[0], names=[ast.alias(name=base_mod_name)])
    ast.copy_location(new_alias_node, node)
    return [new_import_node, new_alias_node]

  def visit_ImportFrom(self, node: ast.ImportFrom):
    # https://docs.python.org/3/library/ast.html#ast.ImportFrom
    if node.level == 0 and not self.mod_map.should_wrap_mod_name(node.module):
      return node
    if node.level > 0:
      return node  # relative imports should already be handled correctly
    assert node.level == 0
    if log.Verbosity >= 10:
      log.unique_print("*** AST transform %r (%s)" % (_ast_get_source_segment(self.src_filename, node), ast.dump(node)))
    new_node = ast.ImportFrom(
      module=self.mod_map.map_mod_name(node.module), names=node.names, level=0)
    ast.copy_location(new_node, node)
    return new_node


def _ast_get_source_segment(src_filename: str, node: ast.AST) -> Optional[str]:
  """Get source code segment of the *source* that generated *node*.

  If some location information (`lineno`, `end_lineno`, `col_offset`,
  or `end_col_offset`) is missing, return None.

  Original ast.get_source_segment is very inefficient. This is faster.
  """
  try:
    lineno = node.lineno - 1
    end_lineno = getattr(node, "end_lineno", node.lineno) - 1
    col_offset = getattr(node, "col_offset", 0)
    end_col_offset = getattr(node, "end_col_offset", -1)
  except AttributeError:
    return None

  lines = linecache.getlines(src_filename)
  if end_lineno == lineno:
    return lines[lineno].encode()[col_offset:end_col_offset].decode()

  first = lines[lineno].encode()[col_offset:].decode()
  last = lines[end_lineno].encode()[:end_col_offset].decode()
  lines = lines[lineno + 1:end_lineno]

  lines.insert(0, first)
  lines.append(last)
  return ''.join(lines)




from __future__ import annotations
import tensorflow as tf
import math
from typing import Optional, Dict, Any
from returnn.tf.layers.basic import BatchNormLayer
from .module import Module
from .utils import _single, _pair, _triple, _reverse_repeat_tuple, _ntuple
from ..common_types import _scalar_or_tuple_any_t, _size_1_t, _size_2_t, _size_3_t
from ..functional import zeros, ones, tensor
from ...tensor import Tensor
from ..parameter import Parameter
from .. import init


class _NormBase(Module):
  """Common base of _InstanceNorm and _BatchNorm"""
  _version = 2

  def __init__(
      self,
      num_features: int,
      eps: float = 1e-5,
      momentum: float = 0.1,
      affine: bool = True,
      track_running_stats: bool = True
  ) -> None:
    super(_NormBase, self).__init__()
    self.num_features = num_features
    self.eps = eps
    self.momentum = momentum
    self.affine = affine
    self.track_running_stats = track_running_stats
    if self.affine:
      self.weight = Parameter(Tensor(num_features))
      self.bias = Parameter(Tensor(num_features))
    else:
      self.register_parameter('weight', None)
      self.register_parameter('bias', None)
    if self.track_running_stats:
      self.register_buffer('running_mean', zeros(num_features))
      self.register_buffer('running_var', ones(num_features))
      self.register_buffer('num_batches_tracked', tensor(0, dtype="int64"))
    else:
      self.register_parameter('running_mean', None)
      self.register_parameter('running_var', None)
      self.register_parameter('num_batches_tracked', None)
    self.reset_parameters()

  def reset_running_stats(self) -> None:
    if self.track_running_stats:
      self.running_mean.zero_()
      self.running_var.fill_(1)
      self.num_batches_tracked.zero_()

  def reset_parameters(self) -> None:
    self.reset_running_stats()
    if self.affine:
      init.ones_(self.weight)
      init.zeros_(self.bias)

  def _check_input_dim(self, input):
    raise NotImplementedError

  def extra_repr(self):
    return '{num_features}, eps={eps}, momentum={momentum}, affine={affine}, ' \
           'track_running_stats={track_running_stats}'.format(**self.__dict__)

  def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                            missing_keys, unexpected_keys, error_msgs):
    version = local_metadata.get('version', None)

    if (version is None or version < 2) and self.track_running_stats:
      # at version 2: added num_batches_tracked buffer
      #               this should have a default value of 0
      num_batches_tracked_key = prefix + 'num_batches_tracked'
      if num_batches_tracked_key not in state_dict:
        state_dict[num_batches_tracked_key] = tensor(0, dtype="int64")

    super(_NormBase, self)._load_from_state_dict(
      state_dict, prefix, local_metadata, strict,
      missing_keys, unexpected_keys, error_msgs)


class _BatchNorm(_NormBase):
  def __init__(self, num_features, eps=1e-5, momentum=0.1, affine=True, track_running_stats=True):
    super(_BatchNorm, self).__init__(num_features, eps, momentum, affine, track_running_stats)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    # Note: Momentum default is quite different. In RETURNN 0.99, in Torch 0.1.
    #  This is not because it's used the other way around.
    # Note: The other default behavior is different as well,
    #  i.e. we need to use `update_sample_only_in_training` and `delay_sample_update`.
    # https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm1d.html
    # https://github.com/pytorch/pytorch/blob/59605811488eb07b3b8bf70a5f0b4b56b34b4a61/aten/src/ATen/native/Normalization.cpp#L546
    return {
      "class": "batch_norm", "from": self._get_input_layer_name(input),
      "update_sample_only_in_training": True, "delay_sample_update": True,
      "param_version": 1,
      "momentum": self.momentum, "epsilon": self.eps}


class BatchNorm1d(_BatchNorm):
  def import_params_torch_to_returnn(self, *, layer: BatchNormLayer, torch_module: BatchNorm1d):
    import numpy
    session = tf.compat.v1.get_default_session()

    # The param names in our RETURNN layer are somewhat strange...
    def _get_param_by_name_postfix(name: str) -> tf.Variable:
      ps = [p for (name_, p) in layer.params.items() if name_.endswith(f"_{name}") or name_.endswith(f"/{name}")]
      assert len(ps) == 1, f"param name {name} not unique or found in layer {layer} with params {layer.params}"
      return ps[0]

    def _expand_dims(x: numpy.ndarray) -> numpy.ndarray:
      assert x.shape == (layer.output.dim,)
      out_shape = [
        layer.output.dim if i == layer.output.feature_dim_axis else 1
        for i in range(layer.output.batch_ndim)]
      return numpy.reshape(x, out_shape)

    def _convert(x: Tensor, name: str):
      tf_param = _get_param_by_name_postfix(name)
      values = _expand_dims(x.detach().cpu().numpy())
      tf_param.load(values, session=session)

    _convert(torch_module.running_mean, "mean")
    _convert(torch_module.running_var, "variance")
    _convert(torch_module.weight, "gamma")
    _convert(torch_module.bias, "beta")


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
import tensorflow as tf
import numpy
from typing import Optional, List, Tuple, Dict
from weakref import ref
from returnn.tf.util.data import Data, DimensionTag
from . import _types
from . import namespace as _namespace
from . import module as _module
from . import call as _call


class TensorEntry:
  tensor: ref[_types.Tensor]
  returnn_data: Optional[Data] = None
  returnn_axis_from_torch_axis: Optional[Dict[int, int]] = None
  validated_to_torch: bool = False
  validated_to_torch_tf_feed_dict: Optional[Dict[tf.Tensor, numpy.ndarray]] = None
  validated_to_torch_tf_sizes_feed_dict: Optional[Dict[tf.Tensor, numpy.ndarray]] = None
  is_param: bool = False
  is_const: bool = False  # e.g. via from_numpy, empty, zeros, etc
  is_input: bool = False  # in TF1 terminology, would be a placeholder
  output_from_modules: List[_module.ModuleEntry]
  output_from_calls: List[_call.CallEntry]
  parent_owning_modules: List[Tuple[_module.ModuleEntry, str]]  # e.g. param or buffer
  creation_stack_call: Optional[_call.CallEntry]
  module_context_stack: List[_module.ModuleEntry]
  names: List[_namespace.RegisteredName]

  def __init__(self, tensor: ref[_types.Tensor],
               creation_stack_call: Optional[_call.CallEntry],
               module_context_stack: List[_module.ModuleEntry]):
    self.tensor = tensor
    self.creation_stack_call = creation_stack_call
    self.module_context_stack = module_context_stack
    self.output_from_modules = []
    self.output_from_calls = []
    self.parent_owning_modules = []
    self.names = []

  def __repr__(self):
    return f"<{self.__class__.__name__} {self.repr_content()}>"

  def repr_content(self):
    if self.returnn_data:
      returnn_data_repr = f"[{','.join(self.returnn_data.get_batch_axes_short_description())}]"
      if self.returnn_axis_from_torch_axis == {i: i for i in range(self.returnn_data.batch_ndim)}:
        mapping_repr = "id"
      else:
        mapping_repr = repr(self.returnn_axis_from_torch_axis).replace(" ", "")
      returnn_data_repr = f"{self.returnn_data.name!r} {returnn_data_repr} axes {mapping_repr}"
    else:
      returnn_data_repr = None
    tensor = self.tensor()
    tensor_repr = repr(tensor.shape).replace(" ", "") if tensor is not None else "?"
    name_repr = self.get_canonical_name(fallback='?')
    if name_repr != "?":
      name_repr = repr(name_repr)
    return (
      f"name:{name_repr} "
      f"tensor:{tensor_repr} "
      f"returnn_data:{returnn_data_repr}")

  def get_canonical_parent_module(self,
                                  parent_namespace: Optional[_namespace.RegisteredName] = None
                                  ) -> Optional[_module.ModuleEntry]:
    if self.parent_owning_modules:
      return self.parent_owning_modules[0][0]
    if self.module_context_stack:
      mod = self.module_context_stack[-1]
      if parent_namespace and mod in parent_namespace.modules:
        pass
      else:
        return mod
    return None

  def get_canonical_name(self, *, fallback: Optional[str] = None) -> str:
    if self.parent_owning_modules:
      return self.parent_owning_modules[0][1]
    if fallback:
      return fallback
    raise NotImplementedError

  def get_returnn_axis_description(self, torch_axis: int) -> str:
    """
    :param torch_axis:
    :return: name such that
      self.returnn_axis_from_torch_axis[self.returnn_data.get_axis_from_description(name)] == torch_axis
    """
    assert self.returnn_data and self.returnn_axis_from_torch_axis is not None
    ndim = self.returnn_data.batch_ndim
    assert -ndim <= torch_axis < ndim
    if torch_axis < 0:
      torch_axis += ndim
    assert 0 <= torch_axis < ndim
    axis = self.returnn_axis_from_torch_axis[torch_axis]
    if axis == self.returnn_data.batch_dim_axis:
      return "B"
    if axis == self.returnn_data.time_dim_axis:
      return "T"
    if axis == self.returnn_data.feature_dim_axis:
      return "F"
    dim_tag = self.returnn_data.get_dim_tag(axis)
    if dim_tag.dyn_size is not None:
      axes_with_same_dim_tag = [ax for ax in range(ndim) if dim_tag == self.returnn_data.get_dim_tag(ax)]
      if len(axes_with_same_dim_tag) == 1:
        return f"stag:{dim_tag.description}"
      else:
        return f"stag-single:{axes_with_same_dim_tag.index(axis) - len(axes_with_same_dim_tag)}:{dim_tag.description}"
    static_axes = self.returnn_data.get_static_axes()
    if axis in static_axes:
      return f"static:{static_axes.index(axis)}"
    spatial_axes = self.returnn_data.get_spatial_batch_axes()
    if axis in spatial_axes:
      return f"spatial:{spatial_axes.index(axis)}"
    raise Exception(f"Should not get here. Dim tag {dim_tag} for axis {axis} for data {self.returnn_data}")

  @property
  def torch_axis_from_returnn_axis(self) -> Dict[int, int]:
    assert self.returnn_data and self.returnn_axis_from_torch_axis is not None
    return {i: j for (j, i) in self.returnn_axis_from_torch_axis.items()}



#!/usr/bin/env python3

"""
Usage example::

  python3 pytorch_to_returnn.py \
  --pwg_config mb_melgan.v2.yaml \
  --pwg_checkpoint mb_melgan_models/checkpoint-1000000steps.pkl \
  --features data/features.npy

"""

import _setup_env  # noqa
import argparse
import torch
import numpy
import yaml
import wave
import better_exchook
import typing
from returnn.tf.util.basic import debug_register_better_repr, setup_tf_thread_pools, print_available_devices
from returnn.log import log


def main():
    parser = argparse.ArgumentParser(description="MB-MelGAN vocoder")
    parser.add_argument("--features", required=True, help="npy file. via decoder.py --dump_features")
    parser.add_argument("--pwg_config", type=str, help="ParallelWaveGAN config (.yaml)")
    parser.add_argument("--pwg_checkpoint", type=str, help="ParallelWaveGAN checkpoint (.pkl)")
    args = parser.parse_args()

    better_exchook.install()
    debug_register_better_repr()
    log.initialize(verbosity=[5])
    setup_tf_thread_pools()
    print_available_devices()

    def model_func(wrapped_import, inputs: torch.Tensor):
        if typing.TYPE_CHECKING or not wrapped_import:
            import torch
            from parallel_wavegan import models as pwg_models
            from parallel_wavegan import layers as pwg_layers

        else:
            torch = wrapped_import("torch")
            wrapped_import("parallel_wavegan")
            pwg_models = wrapped_import("parallel_wavegan.models")
            pwg_layers = wrapped_import("parallel_wavegan.layers")

        # Initialize PWG
        pwg_config = yaml.load(open(args.pwg_config), Loader=yaml.Loader)
        pyt_device = torch.device("cpu")
        generator = pwg_models.MelGANGenerator(**pwg_config['generator_params'])
        generator.load_state_dict(
            torch.load(args.pwg_checkpoint, map_location="cpu")["model"]["generator"])
        generator.remove_weight_norm()
        pwg_model = generator.eval().to(pyt_device)
        assert pwg_config["generator_params"].get("aux_context_window", 0) == 0  # not implemented otherwise
        pwg_pqmf = pwg_layers.PQMF(pwg_config["generator_params"]["out_channels"]).to(pyt_device)

        with torch.no_grad():
            return pwg_pqmf.synthesis(pwg_model(inputs))

    feature_data = numpy.load(args.features)
    print("Feature shape:", feature_data.shape)

    import pytorch_to_returnn.log
    pytorch_to_returnn.log.Verbosity = 6
    from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn
    verify_torch_and_convert_to_returnn(model_func, inputs=feature_data[None, :, :])
    # from pytorch_to_returnn.wrapped_import import wrapped_import_demo
    # from pytorch_to_returnn import torch as torch_returnn
    # model_func(wrapped_import_demo, inputs=torch_returnn.from_numpy(feature_data[None, :, :]))

    audio_waveform = model_func(None, inputs=torch.from_numpy(feature_data[None, :, :]))
    audio_waveform = audio_waveform.view(-1).cpu().numpy()
    audio_raw = numpy.asarray(audio_waveform*(2**15-1), dtype="int16").tobytes()

    out_fn = "out.wav"
    wave_writer = wave.open(out_fn, "wb")
    wave_writer.setnchannels(1)
    wave_writer.setframerate(16000)
    wave_writer.setsampwidth(2)
    wave_writer.writeframes(audio_raw)
    wave_writer.close()
    print("Wrote %s." % out_fn)


if __name__ == "__main__":
    main()




import importlib
import importlib.abc
import importlib.machinery
import sys
import ast
from ... import log
from ..ast_transformer import AstImportTransformer
from ..base_wrappers.module import WrappedModule, WrappedSourceModule, WrappedIndirectModule
from ..context import WrapCtx


# https://docs.python.org/3/library/importlib.html
# https://www.python.org/dev/peps/pep-0302/
# https://dev.to/dangerontheranger/dependency-injection-with-import-hooks-in-python-3-5hap


class MetaPathLoader(importlib.abc.Loader):
  def __init__(self, ctx: WrapCtx):
    self.ctx = ctx
    self.mod_prefix = ctx.wrapped_mod_prefix  # e.g. "pytorch_to_returnn._wrapped_mods."
    assert self.mod_prefix.endswith(".")

  def __repr__(self):
    return "<wrapped mod loader>"

  def create_module(self, spec: importlib.machinery.ModuleSpec) -> WrappedModule:
    if spec.name == self.mod_prefix[:-1]:
      return None  # fall back to default
    assert spec.name.startswith(self.mod_prefix)
    assert spec.name not in sys.modules
    orig_mod_name = spec.name[len(self.mod_prefix):]
    assert not orig_mod_name.startswith(self.mod_prefix)
    # Normal load. This is just to get __file__, and check whether it is correct. Maybe useful otherwise as well.
    orig_mod = importlib.import_module(orig_mod_name)
    assert not isinstance(orig_mod, WrappedModule)
    assert orig_mod.__name__ == orig_mod_name
    orig_mod_name_parts = orig_mod_name.split(".")
    explicit_direct_use = False
    for i in reversed(range(1, len(orig_mod_name_parts) + 1)):
      if ".".join(orig_mod_name_parts[:i]) in self.ctx.wrap_mods_direct:
        explicit_direct_use = True
        break
      if ".".join(orig_mod_name_parts[:i]) in self.ctx.wrap_mods_indirect:
        return WrappedIndirectModule(name=spec.name, orig_mod=orig_mod, ctx=self.ctx)
    orig_mod_loader = orig_mod.__loader__
    if not isinstance(orig_mod_loader, importlib.abc.ExecutionLoader):
      assert not explicit_direct_use
      return WrappedIndirectModule(name=spec.name, orig_mod=orig_mod, ctx=self.ctx)
    src = orig_mod_loader.get_source(orig_mod.__name__)
    if src is None:  # e.g. binary module
      assert not explicit_direct_use
      return WrappedIndirectModule(name=spec.name, orig_mod=orig_mod, ctx=self.ctx)
    return WrappedSourceModule(name=spec.name, orig_mod=orig_mod, source=src, ctx=self.ctx)

  def exec_module(self, module: WrappedModule):
    if module.__name__ == self.mod_prefix[:-1]:  # dummy top-level module
      module.__file__ = None
      module.__loader__ = self
      module.__path__ = []
      module.__package__ = module.__name__
      return
    assert isinstance(module, WrappedModule)
    assert module.__name__.startswith(self.mod_prefix)
    if isinstance(module, WrappedIndirectModule):
      return  # nothing needed to be done
    if log.Verbosity >= 5:
      print("*** exec mod", module)
    assert isinstance(module, WrappedSourceModule)
    # noinspection PyProtectedMember
    orig_mod = module._wrapped__orig_mod
    orig_mod_name = module.__name__[len(self.mod_prefix):]  # e.g. "parallel_wavegan.layers"
    assert orig_mod.__name__ == orig_mod_name
    if not self.ctx.should_wrap_mod(orig_mod_name):
      if log.Verbosity >= 4:
        print(f"*** {self.ctx} extend by mod {orig_mod_name!r}")
      self.ctx.extend_wrap_mod_(orig_mod_name)
    # noinspection PyProtectedMember
    src = module._wrapped__source
    tree = ast.parse(source=src, filename=orig_mod.__file__)
    ast_transformer = AstImportTransformer(
      mod_map=self.ctx.mod_map,
      src_filename=orig_mod.__file__)
    tree = ast_transformer.visit(tree)
    tree = ast.fix_missing_locations(tree)
    code = compile(tree, orig_mod.__file__, "exec")
    is_pkg = getattr(orig_mod, "__path__", None) is not None
    module.__file__ = orig_mod.__file__
    module.__loader__ = self
    if is_pkg:
      module.__path__ = []
      module.__package__ = module.__name__
    else:
      module.__package__ = module.__name__.rpartition('.')[0]
    exec(code, module.__dict__)




import torch
from typing import Optional
from ... import log
from ...naming import Naming


class WrappedModuleBase(torch.nn.Module):
  _wrapped_class_cache = {}  # cls -> WrappedClass

  # Need to overwrite to wrap __init__ to correctly set context.
  def __new__(cls, *args, **kwargs):
    if cls not in cls._wrapped_class_cache:
      class WrappedClass(cls):
        def __init__(self, *args, **kwargs):
          self.__class__ = cls  # we don't need this wrapper class anymore
          if log.Verbosity >= 4:
            log.unique_print(
              "*** torch module create %s.%s(...)" % (self.__class__.__module__, self.__class__.__qualname__))
          with Naming.get_instance().push_module_creation(self):
            cls.__init__(self, *args, **kwargs)
      WrappedClass.__name__ = cls.__name__
      WrappedClass.__qualname__ = cls.__qualname__
      WrappedClass.__module__ = cls.__module__
      wrapped_cls = WrappedClass
      cls._wrapped_class_cache[cls] = wrapped_cls
    else:
      wrapped_cls = cls._wrapped_class_cache[cls]
    return super(WrappedModuleBase, cls).__new__(wrapped_cls)

  def __call__(self, *args, **kwargs):
    if log.Verbosity >= 3:
      log.unique_print(
        "*** torch module call %s.%s(...)(...)" % (self.__class__.__module__, self.__class__.__qualname__))
    with Naming.get_instance().push_module_call(module=self, inputs_args=args, inputs_kwargs=kwargs) as call_entry:
      res = super(WrappedModuleBase, self).__call__(*args, **kwargs)
      call_entry.set_outputs(res)
    return res

  def __setattr__(self, key, value):
    super(WrappedModuleBase, self).__setattr__(key, value)
    if isinstance(value, torch.nn.Module):
      Naming.get_instance().register_module_child_attr(self, key, value)

  def add_module(self, name: str, module: Optional[torch.nn.Module]) -> None:
    super(WrappedModuleBase, self).add_module(name=name, module=module)
    if module:
      Naming.get_instance().register_module_child_attr(self, name, module)

  def get_returnn_name(self) -> str:
    return self.__class__.__name__

  @classmethod
  def has_torch_forward(cls) -> bool:
    in_prefix = False
    from pytorch_to_returnn.import_wrapper.import_ import WrappedModPrefixes
    for prefix in WrappedModPrefixes:
      prefix += "torch.nn."
      if cls.__module__.startswith(prefix):
        in_prefix = True
        break
    if not in_prefix:
      return True
    from pytorch_to_returnn.torch import nn as returnn_torch_nn
    if not hasattr(returnn_torch_nn, cls.__qualname__):
      raise NotImplementedError(f"pytorch_to_returnn.torch.nn.{cls.__qualname__} not yet implemented")
    returnn_cls = getattr(returnn_torch_nn, cls.__qualname__)
    assert issubclass(returnn_cls, returnn_torch_nn.Module)
    return returnn_cls.has_torch_forward()




from typing import Set, Iterable, Dict, Optional, Callable, TypeVar, Type
from .mod_map import ModMap


class WrapCtx:
  def __init__(self, *,
               wrapped_mod_prefix: str,
               wrap_mods_direct: Set[str] = None, wrap_mods_indirect: Set[str] = None,
               wrap_mods_alternatives: Dict[str, str] = None,
               keep_as_is_types: Iterable[type] = (),
               explicit_wrapped_types: Dict[type, "ExplicitWrappedType"] = None):
    """
    :param wrapped_mod_prefix: e.g. "pytorch_to_returnn._wrapped_mods."
    :param wrap_mods_direct: e.g. {"torch.nn.modules"}
      These modules will be imported using the direct AST transformation,
      and result in :class:`WrappedSourceModule`.
    :param wrap_mods_indirect: e.g. {"torch", "torch.nn.modules.module"}
      These modules will be wrapped using :class:`WrappedIndirectModule`,
      which wraps the originally unmodified imported module.
      There can be overlaps between wrap_mods_direct and wrap_mods_indirect.
      The topmost entry will decide whether it is direct or indirect.
    :param wrap_mods_alternatives: e.g. {"torch": "pytorch_to_returnn.torch"}
      These will not be wrapped using :class:`WrappedModule` but can map to some other namespace.
    :param keep_as_is_types: e.g. {torch.device, torch.dtype, torch.Size}
    :param explicit_wrapped_types: e.g. {torch.Tensor: ExplicitWrappedType(WrappedTorchTensor...)}
    """
    assert wrapped_mod_prefix.endswith(".")
    self.wrapped_mod_prefix = wrapped_mod_prefix
    if wrap_mods_direct is None:
      wrap_mods_direct = set()  # type: Set[str]
    self.wrap_mods_direct = wrap_mods_direct
    if wrap_mods_indirect is None:
      wrap_mods_indirect = set()  # type: Set[str]
    self.wrap_mods_indirect = wrap_mods_indirect
    if wrap_mods_alternatives is None:
      wrap_mods_alternatives = {}  # type: Dict[str, str]
    self.wrap_mods_alternatives = wrap_mods_alternatives
    self.keep_as_is_types = tuple(keep_as_is_types)
    if explicit_wrapped_types is None:
      explicit_wrapped_types = {}  # type: Dict[type, "ExplicitWrappedType"]
    self.explicit_wrapped_types = explicit_wrapped_types
    self.mod_map = self._make_mod_map()

  def __repr__(self):
    return "<%s %r>" % (self.__class__.__name__, self.wrapped_mod_prefix[:-1])

  def _make_mod_map(self) -> ModMap:
    d = {
      mod_name: self.wrapped_mod_prefix + mod_name for mod_name in self.wrap_mods_direct | self.wrap_mods_indirect}
    d.update(self.wrap_mods_alternatives)
    mod_map = ModMap(d)
    mod_map.simplify_()
    return mod_map

  def should_wrap_mod(self, mod_name: str) -> bool:
    """
    :param mod_name: e.g. "torch"
    """
    return self.mod_map.should_wrap_mod_name(mod_name)

  def extend_wrap_mod_(self, mod_name: str):
    """
    :param mod_name: e.g. "parallel_wavegan.layers"

    Adds this to wrap_mods_direct if not there already.
    This is an inplace modification.
    """
    if self.should_wrap_mod(mod_name):
      return
    self.wrap_mods_direct.add(mod_name)
    self.mod_map = self._make_mod_map()


class ExplicitWrappedType:
  _OrigType = TypeVar("_OrigType")
  _NewType = TypeVar("_NewType")
  _WrapFuncType = Callable[[_OrigType], _NewType]

  def __init__(self, orig_type: Type[_OrigType], new_type: Type[_NewType], wrap: Optional[_WrapFuncType]):
    self.orig_type = orig_type
    self.new_type = new_type
    self._wrap = wrap

  def wrap(self, obj: _OrigType, *, name: str, ctx: WrapCtx) -> _NewType:
    if self._wrap:
      return self._wrap(obj)
    return self.new_type(orig_obj=obj, name=name, ctx=ctx)


def make_torch_traced_ctx(wrapped_mod_prefix: str) -> WrapCtx:
  """
  This wraps `torch` imports to use mostly the original `torch` module
  but installs some tracing wrappers,
  which keep track of the tensors, and create names using the :class:`Naming` logic.
  """
  import torch
  from .torch_wrappers import WrappedTorchTensor, WrappedTorchParameter, WrappedModuleBase

  _KeepAsIsTypes = (
    torch.device,
    torch.Size,
    torch.dtype,
  )

  def _wrap_tensor(obj):
    assert isinstance(obj, torch.Tensor)
    obj = obj.as_subclass(WrappedTorchTensor)
    return obj

  # noinspection PyUnusedLocal
  def _raise_not_implemented(*args):
    raise NotImplementedError

  _TypeMap = {
    torch.Tensor: ExplicitWrappedType(torch.Tensor, WrappedTorchTensor, wrap=_wrap_tensor),
    torch.nn.Parameter: ExplicitWrappedType(torch.nn.Parameter, WrappedTorchParameter, wrap=_raise_not_implemented),
    torch.nn.Module: ExplicitWrappedType(torch.nn.Module, WrappedModuleBase, wrap=_raise_not_implemented),
  }

  return WrapCtx(
    wrapped_mod_prefix=wrapped_mod_prefix,
    wrap_mods_direct=_TorchExplicitDirectModList,
    wrap_mods_indirect=_TorchExplicitIndirectModList,
    keep_as_is_types=_KeepAsIsTypes,
    explicit_wrapped_types=_TypeMap)


_TorchExplicitIndirectModList = {
  # Leave "torch" in to not do any magic on the original Torch modules.
  # Note that the main torch module cannot be transformed
  # because of some side-effects triggered by native Torch code,
  # which we cannot catch automatically.
  # More specifically, in torch/__init__.py, there is the call _C._initExtension,
  # which (un-intuitively) does all sorts of initialization,
  # including torch::utils::initializeDtypes,
  # which itself has the side-effect of adding dtypes (e.g. `float32`)
  # to the `torch` module (-> `torch.float32`).
  # And the name `"torch"` is hardcoded there,
  # i.e. it will not used our wrapped module,
  # thus we will never have `float32` registered,
  # and that leads to exceptions.
  # This is likely not the only problem.
  "torch",
}
if "torch" in _TorchExplicitIndirectModList:
  _TorchExplicitIndirectModList.update({
    # If we would transform `torch` directly, the following should be explicitly excluded.
    "torch.tensor", "torch.distributed",
    "torch._C",
    # Avoid to have the `Module` class twice.
    "torch.nn.modules.module",
  })

_TorchExplicitDirectModList = {
  # Note: When we don't wrap all of `torch` (i.e. `"torch"` is in _ExplicitIndirectModList),
  # this will partially be WrappedIndirectModule and WrappedSourceModule.
  # This is not a problem in principle, but there can be subtle problems.
  # Also the import order can be important.
  # E.g. "torch.nn.modules" should be imported before "torch.nn.functional".
  # This might need extra explicit handling, depending what we have here in this list.
  "torch.nn.modules",
  # "torch.nn.functional",  # -- not needed, also causes other problems
}


def make_torch_returnn_ctx(wrapped_mod_prefix: str) -> WrapCtx:
  """
  This wraps `torch` imports to use our `python_to_returnn.torch` module instead.
  """
  from pytorch_to_returnn import torch as torch_returnn
  return WrapCtx(
    wrapped_mod_prefix=wrapped_mod_prefix,
    wrap_mods_alternatives={"torch": torch_returnn.__package__})




import importlib
import importlib.abc
import importlib.machinery
import types
from typing import Optional
from .loader import MetaPathLoader


# https://docs.python.org/3/library/importlib.html
# https://www.python.org/dev/peps/pep-0302/
# https://dev.to/dangerontheranger/dependency-injection-with-import-hooks-in-python-3-5hap


class MetaPathFinder(importlib.abc.MetaPathFinder):
  def __init__(self, loader: MetaPathLoader):
    self.loader = loader
    self.ctx = loader.ctx

  def find_spec(self, fullname: str, path: Optional[str], target: Optional[types.ModuleType] = None):
    if fullname.startswith(self.loader.mod_prefix) or fullname == self.loader.mod_prefix[:-1]:
      return importlib.machinery.ModuleSpec(fullname, self.loader)
    return None




from __future__ import annotations
from tensorflow.python.util import nest
from typing import Generator, List, Optional, Union, Dict, Any, Tuple
from contextlib import contextmanager
from collections import OrderedDict
from weakref import ref, WeakKeyDictionary
import numpy
from returnn.tf.util.data import Data
from returnn.tf.layers.basic import LayerBase
from . import _types
from . import tensor as _tensor
from . import module as _module
from . import call as _call
from . import namespace as _namespace


class Naming:
  tensors: WeakKeyDictionary[_types.Tensor, _tensor.TensorEntry]
  const_tensor_cache: List[_types.Tensor]
  modules: OrderedDict[_types.Module, _module.ModuleEntry]
  inputs: List[_tensor.TensorEntry]
  outputs: List[_tensor.TensorEntry]
  module_creation_stack: List[_module.ModuleEntry]
  module_apply_stack: List[_module.ModuleEntry]
  module_context_stack: List[_module.ModuleEntry]
  module_call_stack: List[_call.CallEntry]
  root_func_calls: List[_call.CallEntry]
  _default_instance: Optional[Naming] = None
  _instance_stack: List[Naming] = []

  @classmethod
  @contextmanager
  def make_instance(cls, **kwargs) -> Generator[Naming]:
    ctx = Naming(**kwargs)
    cls._instance_stack.append(ctx)
    yield ctx
    assert cls._instance_stack[-1] is ctx
    cls._instance_stack.pop(-1)

  @classmethod
  def get_instance(cls) -> Naming:
    if cls._instance_stack:
      return cls._instance_stack[-1]
    if not cls._default_instance:
      cls._default_instance = Naming()
    return cls._default_instance

  def __init__(self, *,
               wrap_to_returnn_enabled: bool = True,
               returnn_train_flag: bool = False,
               keep_orig_module_io_tensors: bool = True,
               import_params_from_torch_namespace: Optional[Naming] = None,
               validate_allclose_kwargs: Optional[Dict[str, Any]] = None,
               ):
    """
    :param wrap_to_returnn_enabled: Will construct corresponding RETURNN layers.
    :param returnn_train_flag: set RETURNN TFNetwork train_flag accordingly
    :param keep_orig_module_io_tensors: Keeps references to the original (or wrapped) torch.Tensor instances
       for all module calls. This will need extra memory.
    :param import_params_from_torch_namespace: If given, we try to import params.
    :param validate_allclose_kwargs: for numpy.allclose
    """
    self.wrap_to_returnn_enabled = wrap_to_returnn_enabled
    self.returnn_train_flag = returnn_train_flag
    self.keep_orig_module_io_tensors = keep_orig_module_io_tensors
    self.import_params_from_torch_namespace = import_params_from_torch_namespace
    if validate_allclose_kwargs is None:
      # PyTorch uses some different algos, e.g. different convolution,
      # which leads to quite huge relative differences (for values close to 0.0).
      validate_allclose_kwargs = dict(rtol=0, atol=5e-4)
    self.validate_allclose_kwargs = validate_allclose_kwargs
    self.tensors = WeakKeyDictionary()
    self.const_tensor_cache = []
    self.modules = OrderedDict()
    self.inputs = []
    self.outputs = []
    self.module_context_stack = []
    self.module_creation_stack = []
    self.module_apply_stack = []
    self.module_call_stack = []
    self.root_func_calls = []
    self.root_namespace = _namespace.RegisteredName(
      parent=None,
      wrap_to_returnn_enabled=wrap_to_returnn_enabled,
      returnn_train_flag=returnn_train_flag,
      is_subnet=True)
    self.tmp_eager_root_namespace = self.root_namespace.register_sub_net(suggested_name=".tmp_root")

  @contextmanager
  def push_module_creation(self, module: _types.Module) -> _module.ModuleEntry:
    assert module not in self.modules
    entry = _module.ModuleEntry(module=module)
    self.modules[module] = entry
    self.module_creation_stack.append(entry)
    with self.push_module_context(module):
      yield entry
    assert self.module_creation_stack[-1] is entry
    self.module_creation_stack.pop(-1)

  @contextmanager
  def push_module_apply(self, module: _types.Module) -> _module.ModuleEntry:
    entry = self.modules[module]
    self.module_apply_stack.append(entry)
    with self.push_module_context(module):
      yield entry
    assert self.module_apply_stack[-1] is entry
    self.module_apply_stack.pop(-1)

  def push_module_context(self, module: _types.Module) -> _module.ModuleEntry:
    entry = self.modules[module]
    if self.module_context_stack:
      prev_top = self.module_context_stack[-1]
      if prev_top not in entry.parent_context_modules and prev_top.module is not module:
        entry.parent_context_modules.append(prev_top)
    self.module_context_stack.append(entry)
    return entry

  def pop_module_context(self, module: _types.Module):
    entry = self.modules[module]
    assert self.module_context_stack[-1] is entry
    self.module_context_stack.pop(-1)

  def prepare_tensor_as_input(self, x: _tensor.TensorEntry, *, parent_namespace: _namespace.RegisteredName):
    names = [name_ for name_ in x.names if parent_namespace in name_.get_parents_hierarchy()]
    if names:
      return  # ok
    if x.is_param:
      # This should be via the `Parameter` class. We don't really enforce this here, though.
      # Earlier, we also asserted that returnn_data must be set and its placeholder is not set yet.
      # We don't enforce this anymore. If the variable was created before in the temp namespace,
      # and now accessed in the real namespace, we want to create it again.
      from pytorch_to_returnn.torch.nn.modules import Variable
      param_name = x.get_canonical_name()
      if x.returnn_data and x.returnn_data.name == "_unnamed_param":
        x.returnn_data.name = f"param:{param_name}"
      parent_mod = x.get_canonical_parent_module()
      prefix = (parent_mod.get_canonical_name() + "_") if parent_mod else ""
      mod = Variable(param=x.tensor(), parent_mod=(parent_mod.module, param_name))
      self.modules[mod].canonical_name = prefix + param_name
      res = mod()
      res_tensor = self.tensors[res]
      assert isinstance(res_tensor, _tensor.TensorEntry)
      assert res_tensor.returnn_data.placeholder is not None
      if x.returnn_data:
        x.returnn_data.placeholder = res_tensor.returnn_data.placeholder
    elif not x.output_from_calls or x.is_const:
      # Assume this is a constant.
      x.is_const = True
      const_name = x.get_canonical_name(fallback="unnamed_const")
      tensor = x.tensor()
      if not x.returnn_data:
        x.returnn_data = Data(
          name=f"const:{const_name}", shape=tensor.shape, dtype=tensor.dtype.name,
          batch_dim_axis=None, time_dim_axis=None)
        x.returnn_axis_from_torch_axis = {i: i for i in range(len(tensor.shape))}
      parent_mod = x.get_canonical_parent_module()
      prefix = (parent_mod.get_canonical_name() + "_") if parent_mod else ""
      from pytorch_to_returnn.torch.nn.modules import Constant
      mod = Constant(value=tensor)
      self.modules[mod].canonical_name = prefix + const_name
      res = mod()
      res_tensor = self.tensors[res]
      assert isinstance(res_tensor, _tensor.TensorEntry)
      assert res_tensor.returnn_data.placeholder is not None
      x.returnn_data.placeholder = res_tensor.returnn_data.placeholder
    else:
      raise Exception(f"Cannot handle tensor {x}, via {x.output_from_calls} ...")

  def _prepare_module_call_returnn_inputs(self, call: _call.CallEntry):
    """
    It means this module has no forward, i.e. it is wrapped as RETURNN layer.
    We might need to make some inputs available, which are not available yet,
    e.g. constants, params, etc.
    """
    if not self.wrap_to_returnn_enabled:
      return
    call_parent_namespace = call.namespace.parent or self.root_namespace
    for x in call.inputs_flat:
      if not isinstance(x, _tensor.TensorEntry):
        continue
      self.prepare_tensor_as_input(x, parent_namespace=call_parent_namespace)

  def push_module_call(self, *, module: _types.Module,
                       inputs_args: Tuple[Union[_types.Tensor, Any], ...],
                       inputs_kwargs: Dict[str, Union[_types.Tensor, Any]]) -> _call.CallEntry:
    module_entry = self.modules[module]
    assert isinstance(module_entry, _module.ModuleEntry)
    entry = _call.CallEntry(module=module_entry)
    inputs_flat = nest.flatten((inputs_args, inputs_kwargs))
    if self.keep_orig_module_io_tensors:
      entry.orig_inputs_args = inputs_args
      entry.orig_inputs_kwargs = inputs_kwargs
      entry.orig_inputs_flat = inputs_flat
    entry.inputs_flat = [self._make_tensor(x) for x in inputs_flat]
    if self.wrap_to_returnn_enabled:
      entry.inputs_args, entry.inputs_kwargs = nest.pack_sequence_as(
        structure=(inputs_args, inputs_kwargs), flat_sequence=entry.inputs_flat)
    entry.level = len(self.module_call_stack)
    if self.module_call_stack:
      recent_entry = self.module_call_stack[-1]
      recent_entry.child_calls.append(entry)
      entry.parent_call = recent_entry
    else:
      self.root_func_calls.append(entry)
    if self.module_creation_stack or self.module_apply_stack:
      # This is likely for parameter initialization, or setting up constants.
      # Keep this in a separate namespace.
      root_namespace = self.tmp_eager_root_namespace
    else:
      root_namespace = self.root_namespace
    if entry.parent_call:
      assert entry.parent_call.namespace
      parent_namespace = entry.parent_call.namespace
      # Get the parent namespace, which is a subnetwork.
      while not parent_namespace.is_subnetwork():
        assert parent_namespace.parent
        parent_namespace = parent_namespace.parent
    else:
      parent_namespace = root_namespace
    assert parent_namespace.is_subnetwork()
    if parent_namespace is root_namespace and _flatten_namespace_for_mod(module_entry):
      # Special case, somewhat nicer to flatten the namespace for this case.
      root_namespace.assign_call(entry)
      namespace = root_namespace
    else:
      namespace = parent_namespace.register_sub_call(entry)
    assert module_entry in namespace.modules

    if module.has_torch_forward():
      # This will get an own subnet, and we might create constants,
      # so make sure we create them in the right (parent) namespace.
      self._prepare_module_call_returnn_inputs(entry)

    self.module_call_stack.append(entry)
    assert entry.namespace
    if not module.has_torch_forward():  # no subnet, so do now, to have better namings
      self._prepare_module_call_returnn_inputs(entry)
    return entry

  def pop_module_call(self, call: _call.CallEntry):
    assert self.module_call_stack[-1] is call
    self.module_call_stack.pop(-1)

  def register_module_child_attr(self, parent: _types.Module, attr: str, child: Union[_types.Module, _types.Tensor]):
    assert getattr(parent, attr) is child
    parent_entry = self.modules[parent]
    if child in self.modules:
      child_entry = self.modules[child]
      assert isinstance(child_entry, _module.ModuleEntry)
    else:
      assert child in self.tensors
      child_entry = self.tensors[child]
      assert isinstance(child_entry, _tensor.TensorEntry)
      for parent_param in parent.parameters(recurse=False):
        if parent_param is child:
          child_entry.is_param = True
          break
    if (parent_entry, attr) not in child_entry.parent_owning_modules:
      child_entry.parent_owning_modules.append((parent_entry, attr))

  def register_tensor(self, tensor: _types.Tensor) -> _tensor.TensorEntry:
    if tensor not in self.tensors:
      self.tensors[tensor] = _tensor.TensorEntry(
        tensor=ref(tensor),
        creation_stack_call=self.module_call_stack[-1] if self.module_call_stack else None,
        module_context_stack=list(self.module_context_stack))
    return self.tensors[tensor]

  def get_tensor(self, x: Union[_types.Tensor, int, float, numpy.number, numpy.ndarray, Any]) -> _tensor.TensorEntry:
    """
    Call this when you expect the input to be a tensor.
    If it is a const Python number (int or so), it would create a corresponding const tensor.
    """
    from pytorch_to_returnn.torch import Tensor
    if isinstance(x, Tensor):
      return self.tensors[x]  # we expect this to be registered
    return self.make_const_tensor(x)

  def make_const_tensor(self, x: Union[int, float, numpy.number, numpy.ndarray, Any]) -> _tensor.TensorEntry:
    from pytorch_to_returnn.torch import from_numpy, Tensor
    assert isinstance(x, (int, float, bool, numpy.number, numpy.ndarray))
    x = from_numpy(x)
    assert isinstance(x, Tensor)
    self.const_tensor_cache.append(x)
    return self.tensors[x]

  def _make_tensor(self, x: Union[_types.Tensor, int, float, numpy.number, numpy.ndarray, Any]
                   ) -> Optional[_tensor.TensorEntry]:
    """
    We have to decide which objects we keep as-is (as constant),
    and which we convert to a TensorEntry.
    This is somewhat arbitrary, but should in principle not matter.
    """
    if x is None:
      return None
    if isinstance(x, (int, float, bool, numpy.number)):
      return x  # keep as-is for now. would be handled via `get_tensor` later on
    if not self.wrap_to_returnn_enabled:
      if x in self.tensors:
        return self.tensors[x]
      return None
    from pytorch_to_returnn.torch import Tensor
    if isinstance(x, (numpy.ndarray,)):
      return self.make_const_tensor(x)
    assert isinstance(x, Tensor)
    return self.tensors[x]

  def register_input(self, tensor: _types.Tensor, returnn_data: Data) -> Data:
    from pytorch_to_returnn.torch import Tensor
    assert isinstance(tensor, Tensor)
    entry = self.register_tensor(tensor)
    entry.is_input = True
    entry.is_const = False
    self.inputs.append(entry)
    assert tensor.dim() == returnn_data.batch_ndim
    assert all([dim in {tensor.shape[i], None} for i, dim in enumerate(returnn_data.batch_shape)])
    entry.returnn_data = Data(
      name=returnn_data.name, auto_create_placeholders=True,
      sparse=returnn_data.sparse,
      dim=returnn_data.dim,
      shape=returnn_data.shape,
      batch_dim_axis=returnn_data.batch_dim_axis,
      time_dim_axis=returnn_data.time_dim_axis,
      feature_dim_axis=returnn_data.feature_dim_axis_or_unspecified,
      available_for_inference=True)
    entry.returnn_axis_from_torch_axis = {i: i for i in range(returnn_data.batch_ndim)}
    if entry.returnn_data.have_batch_axis():
      tensor.shape[entry.returnn_data.batch_dim_axis].is_batch_dim = True
    self.root_namespace.register_input(tensor=entry)
    assert entry.returnn_data
    return entry.returnn_data

  def register_output(self, tensor: _types.Tensor) -> _tensor.TensorEntry:
    assert tensor in self.tensors
    entry = self.tensors[tensor]
    assert isinstance(entry, _tensor.TensorEntry)
    assert not entry.is_param and not entry.is_const and not entry.is_input  # not implemented, although simple...
    self.outputs.append(entry)
    if self.wrap_to_returnn_enabled:
      self.root_namespace.register_returnn_subnet_output(entry)
    return entry

  def get_module_abs_call_name(self, module: _types.Module) -> str:
    mod_entry = self.modules[module]
    assert mod_entry.calls  # otherwise it would not have a name
    call = mod_entry.calls[0]
    return call.namespace.get_absolute_name()

  def get_module_by_abs_call_name(self, name: str) -> _types.Module:
    namespace = self.root_namespace
    if name:
      for part_name in name.split("/"):
        if part_name[:1].isnumeric():
          part_name = f"layer{part_name}"
        namespace = namespace.childs_by_name[part_name]
    assert namespace.modules
    return namespace.modules[0].module

  def get_module_abs_name(self, module: _types.Module,
                          root_namespace: Optional[_namespace.RegisteredName] = None) -> Optional[str]:
    """
    Returns an attrib chain, such that `root_mod.attrib1.attrib2...` resolves to `module`,
    for some `root_mod` (in root_namespace.modules).
    Specifically, it would return `attrib1.attrib2...` in that case.
    The inverse is :func:`get_module_by_abs_name`.
    """
    parts = []
    mod_entry = self.modules[module]
    if not root_namespace:
      root_namespace = self.root_namespace
    while True:
      if mod_entry in root_namespace.modules:
        break
      if mod_entry.parent_owning_modules:
        mod_entry, name = mod_entry.parent_owning_modules[0]
        parts.append(name)
        continue
      return None
    abs_name = ".".join(reversed(parts))
    checked_mod = self.get_module_by_abs_name(abs_name)
    assert checked_mod is module
    return abs_name

  def get_module_by_abs_name(self, name: str) -> _types.Module:
    """
    This gets the module by the attrib chain,
    starting from the root modules (root_namespace.modules).
    This is the inverse of :func:`get_module_abs_name`.

    Note that this is different from the registered name hierarchy (via root_namespace).
    The registered name hierarchy reflects the call hierarchy,
    and corresponding canonical names from a call.
    """
    potential_modules = [m.module for m in self.root_namespace.modules]
    if name:
      name_parts = name.split(".")
      for i, part_name in enumerate(name_parts):
        possible_modules = [m for m in potential_modules if hasattr(m, part_name)]
        assert possible_modules, f"{potential_modules} have not {part_name!r}, after {'.'.join(name_parts[:i + 1])!r}"
        potential_modules = [getattr(m, part_name) for m in possible_modules]
    assert len(potential_modules) == 1, f"unique mod for name {name!r}"
    return potential_modules[0]

  def get_module_abs_id_name(self, module: _types.Module) -> str:
    """
    This is a mixture of :func:`get_module_abs_name` and :func:`get_module_abs_call_name`.
    Not all modules are accessible via an attrib chain,
    e.g. when they are temporarily created and not assigned to an attrib.
    Here, we try to resolve an attrib chain, but fall back to call names if not possible.
    The inverse is :func:`get_module_by_abs_id_name`.
    """
    parts = []
    mod_entry = self.modules[module]
    visited = set()
    while True:
      assert mod_entry not in visited
      visited.add(mod_entry)
      if mod_entry in self.root_namespace.modules:
        break
      # Use get_module_abs_name to check whether we have a unique name directly up to the root namespace.
      # If not, directly go to the fallback below via the call name.
      if self.get_module_abs_name(mod_entry.module) is not None:
        assert mod_entry.parent_owning_modules
        mod_entry, name = mod_entry.parent_owning_modules[0]
        parts.append(name)
        continue
      # Fallback to call hierarchy.
      assert mod_entry.names
      name_ = mod_entry.names[0]
      parts.append(f"({name_.name})")
      if name_.parent is self.root_namespace:
        break
      assert name_.parent and name_.parent.modules
      mod_entry = name_.parent.modules[0]
    abs_name = ".".join(reversed(parts))
    checked_mod = self.get_module_by_abs_id_name(abs_name)
    assert checked_mod is module
    return abs_name

  def get_module_by_abs_id_name(self, name: str) -> _types.Module:
    """
    This is the inverse of :func:`get_module_abs_id_name`.
    """
    potential_namespaces = [self.root_namespace]
    potential_modules = [m for m in self.root_namespace.modules]
    if name:
      name_parts = name.split(".")
      for i, part_name in enumerate(name_parts):
        if part_name[:1] == "(" and part_name[-1:] == ")":
          part_name = part_name[1:-1]
          possible_namespaces = [name_ for name_ in potential_namespaces if part_name in name_.childs_by_name]
          assert possible_namespaces, (
            f"{potential_namespaces} have not {part_name!r}, after {'.'.join(name_parts[:i + 1])!r}")
          potential_namespaces = [name_.childs_by_name[part_name] for name_ in potential_namespaces]
          potential_modules = []
          for name_ in potential_namespaces:
            for m in name_.modules:
              if m not in potential_modules:
                potential_modules.append(m)
        else:
          possible_modules = [m for m in potential_modules if hasattr(m.module, part_name)]
          assert possible_modules, f"{potential_modules} have not {part_name!r}, after {'.'.join(name_parts[:i + 1])!r}"
          potential_modules = [self.modules[getattr(m.module, part_name)] for m in possible_modules]
          potential_namespaces = []
          for m in potential_modules:
            for name_ in m.names:
              if name_ not in potential_namespaces:
                potential_namespaces.append(name_)
    assert len(potential_modules) == 1, f"unique mod for name {name!r}"
    return potential_modules[0].module

  def get_module_call_idx(self, *, module: _types.Module, call: _call.CallEntry) -> int:
    mod_entry = self.modules[module]
    assert call in mod_entry.calls
    return mod_entry.calls.index(call)

  def get_module_calls(self, module: _types.Module) -> List[_call.CallEntry]:
    return self.modules[module].calls

  def get_root_module_calls(self) -> OrderedDict[str, _call.CallEntry]:
    d = OrderedDict()
    for name, sub in self.root_namespace.childs_by_name.items():
      if sub.calls:
        d[name] = sub.calls[0]
    return d

  def get_modules_with_params_by_abs_name(self) -> OrderedDict[str, _types.Module]:
    d = OrderedDict()
    _visited = set()

    def visit(namespace: _namespace.RegisteredName, prefix: str):
      for mod in namespace.modules:
        if mod.module in _visited:
          return
        if not list(mod.module.parameters(recurse=False)):
          continue
        assert (prefix == "" or prefix.endswith(".")) and len(namespace.modules) == 1
        d[prefix[:-1]] = mod.module
        _visited.add(mod.module)
      for name, sub in namespace.childs_by_name.items():
        if name.startswith("."):
          continue  # skip hidden sub spaces
        visit(namespace=sub, prefix=f"{prefix}{name}.")

    visit(namespace=self.root_namespace, prefix="")
    return d

  def get_returnn_layer_from_module(self, module: _types.Module) -> LayerBase:
    assert self.wrap_to_returnn_enabled
    entry = self.modules[module]
    assert entry.calls
    call = entry.calls[0]
    assert call.returnn_layer
    return call.returnn_layer


def _flatten_namespace_for_mod(mod_entry: _module.ModuleEntry) -> bool:
  if mod_entry.parent_owning_modules:
    # Use the parent module.
    return False
  mod = mod_entry.module
  if not mod.has_torch_forward():
    # For RETURNN wrapped modules, e.g. it means that it has no forward, but wraps to a RETURNN layer.
    # So, keep this as a separate item, do not flatten it.
    return False
  return True




import warnings
from typing import Dict, Any, Optional
from .module import Module
from ...tensor import Tensor


class _ActivationReturnn(Module):
  func_name: str

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "activation", "activation": self.func_name, "from": self._get_input_layer_name(input)}


class ReLU(_ActivationReturnn):
  func_name = "relu"


class GELU(_ActivationReturnn):
  func_name = "gelu3"


class PReLU(_ActivationReturnn):
  pass


class ELU(_ActivationReturnn):
  pass


class Abs(_ActivationReturnn):
  func_name = "abs"


class Sqrt(_ActivationReturnn):
  func_name = "sqrt"


class Tanh(_ActivationReturnn):
  func_name = "tanh"


class Log(_ActivationReturnn):
  func_name = "log"


class Sigmoid(_ActivationReturnn):
  func_name = "sigmoid"


class LogSigmoid(_ActivationReturnn):
  func_name = "log_sigmoid"


class Power(Module):
  is_original_torch_module = False

  def __init__(self, exponent: float) -> None:
    super(Power, self).__init__()
    self.exponent = exponent

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "eval", "eval": f"tf.math.pow(source(0), {self.exponent})",
      "from": self._get_input_layer_name(input)}


class LeakyReLU(Module):
  def __init__(self, negative_slope: float = 1e-2, inplace: bool = False) -> None:
    super(LeakyReLU, self).__init__()
    self.negative_slope = negative_slope
    assert not inplace  # not supported/implemented -- see :doc:`Unsupported`

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "eval", "eval": f"tf.nn.leaky_relu(source(0), alpha={self.negative_slope})",
      "from": self._get_input_layer_name(input)}


class Softmax(Module):
  _name = "softmax"

  def __init__(self, dim: Optional[int] = None) -> None:
    super(Softmax, self).__init__()
    self.dim = dim

  @classmethod
  def _get_default_softmax_dim(cls, *, ndim: int) -> int:
    warnings.warn(f"Implicit dimension choice for {cls._name} has been deprecated. "
                  "Change the call to include dim=X as an argument.", stacklevel=2)
    if ndim == 0 or ndim == 1 or ndim == 3:
      return 0
    else:
      return 1

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    if self.dim is not None:
      dim = self.dim
    else:
      dim = self._get_default_softmax_dim(ndim=input.ndim)
    returnn_axis = self._get_input_axis_to_returnn(input, dim)
    if returnn_axis == "F":
      return {"class": "activation", "activation": self._name, "from": self._get_input_layer_name(input)}
    return {
      "class": "softmax_over_spatial", "axis": returnn_axis, "from": self._get_input_layer_name(input),
      "log_space": {"softmax": False, "log_softmax": True}[self._name]}


class LogSoftmax(Softmax):
  _name = "log_softmax"


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




"""
Alternative to the original pprint module.
This one has different behavior for indentation, specifically for dicts.
Also the order of dict items are kept as-is
(which is fine for newer Python versions, which will be the insertion order).

Compare (via our ``pprint``)::

  {
    'melgan': {
      'class': 'subnetwork',
      'from': 'data',
      'subnetwork': {
        'l0': {'class': 'pad', 'mode': 'reflect', 'axes': 'spatial', 'padding': (3, 3), 'from': 'data'},
        'la1': {
          'class': 'conv',
          'from': 'l0',
          'activation': None,
          'with_bias': True,
          'n_out': 384,
          'filter_size': (7,),
          'padding': 'valid',
          'strides': (1,),
          'dilation_rate': (1,)
        },
        'lay2': {'class': 'eval', 'eval': 'tf.nn.leaky_relu(source(0), alpha=0.2)', 'from': 'la1'},
        'layer3_xxx': {
          'class': 'transposed_conv',
          'from': 'lay2',
          'activation': None,
          'with_bias': True,
          'n_out': 192,
          'filter_size': (10,),
          'strides': (5,),
          'padding': 'valid',
          'output_padding': (1,),
          'remove_padding': (3,)
        },
        'output': {'class': 'copy', 'from': 'layer3_xxx'}
      }
    },
    'output': {'class': 'copy', 'from': 'melgan'}
  }

Vs (via original ``pprint``)::

  {'melgan': {'class': 'subnetwork',
              'from': 'data',
              'subnetwork': {'l0': {'axes': 'spatial',
                                    'class': 'pad',
                                    'from': 'data',
                                    'mode': 'reflect',
                                    'padding': (3, 3)},
                             'la1': {'activation': None,
                                     'class': 'conv',
                                     'dilation_rate': (1,),
                                     'filter_size': (7,),
                                     'from': 'l0',
                                     'n_out': 384,
                                     'padding': 'valid',
                                     'strides': (1,),
                                     'with_bias': True},
                             'lay2': {'class': 'eval',
                                      'eval': 'tf.nn.leaky_relu(source(0), '
                                              'alpha=0.2)',
                                      'from': 'la1'},
                             'layer3_xxx': {'activation': None,
                                            'class': 'transposed_conv',
                                            'filter_size': (10,),
                                            'from': 'lay2',
                                            'n_out': 192,
                                            'output_padding': (1,),
                                            'padding': 'valid',
                                            'remove_padding': (3,),
                                            'strides': (5,),
                                            'with_bias': True},
                             'output': {'class': 'copy', 'from': 'layer3_xxx'}}},
   'output': {'class': 'copy', 'from': 'melgan'}}

This is a very simple implementation.
There are other similar alternatives:

* [Rich](https://github.com/willmcgugan/rich)
* [pprint++](https://github.com/wolever/pprintpp)

"""

from typing import Any
import sys
import numpy


def pprint(o: Any, *, file=sys.stdout,
           prefix="", postfix="",
           line_prefix="", line_postfix="\n") -> None:
  if "\n" in line_postfix and _type_simplicity_score(o) <= _type_simplicity_limit:
    prefix = f"{line_prefix}{prefix}"
    line_prefix = ""
    postfix = postfix + line_postfix
    line_postfix = ""

  def _sub_pprint(o: Any, prefix="", postfix="", inc_indent=True):
    multi_line = "\n" in line_postfix
    if not multi_line and postfix.endswith(","):
      postfix += " "
    pprint(
      o, file=file, prefix=prefix, postfix=postfix,
      line_prefix=(line_prefix + "  " * inc_indent) if multi_line else "",
      line_postfix=line_postfix)

  def _print(s: str, is_end: bool = False):
    nonlocal prefix  # no need for is_begin, just reset prefix
    file.write(line_prefix)
    file.write(prefix)
    file.write(s)
    if is_end:
      file.write(postfix)
    file.write(line_postfix)
    if "\n" in line_postfix:
      file.flush()
    prefix = ""

  def _print_list():
    for i, v in enumerate(o):
      _sub_pprint(v, postfix="," if i < len(o) - 1 else "")

  if isinstance(o, list):
    if len(o) == 0:
      _print("[]", is_end=True)
      return
    _print("[")
    _print_list()
    _print("]", is_end=True)
    return

  if isinstance(o, tuple):
    if len(o) == 0:
      _print("()", is_end=True)
      return
    if len(o) == 1:
      _sub_pprint(o[0], prefix=f"{prefix}(", postfix=f",){postfix}", inc_indent=False)
      return
    _print("(")
    _print_list()
    _print(")", is_end=True)
    return

  if isinstance(o, set):
    if len(o) == 0:
      _print("set()", is_end=True)
      return
    _print("{")
    _print_list()
    _print("}", is_end=True)
    return

  if isinstance(o, dict):
    if len(o) == 0:
      _print("{}", is_end=True)
      return
    _print("{")
    for i, (k, v) in enumerate(o.items()):
      _sub_pprint(v, prefix=f"{k!r}: ", postfix="," if i < len(o) - 1 else "")
    _print("}", is_end=True)
    return

  if isinstance(o, numpy.ndarray):
    _sub_pprint(
      o.tolist(),
      prefix=f"{prefix}numpy.array(",
      postfix=f", dtype=numpy.{o.dtype}){postfix}",
      inc_indent=False)
    return

  # fallback
  _print(repr(o), is_end=True)


def pformat(o: Any) -> str:
  import io
  s = io.StringIO()
  pprint(o, file=s)
  return s.getvalue()


_type_simplicity_limit = 120.  # magic number


def _type_simplicity_score(o: Any, _offset=0.) -> float:
  """
  :param Any o:
  :param float _offset:
  :return: a score, which is a very rough estimate of len(repr(o)), calculated efficiently
  """
  _spacing = 2.
  if isinstance(o, bool):
    return 4. + _offset
  if isinstance(o, (int, numpy.integer)):
    if o == 0:
      return 1. + _offset
    return 1. + numpy.log10(abs(o)) + _offset
  if isinstance(o, str):
    return 2. + len(o) + _offset
  if isinstance(o, (float, complex, numpy.number)):
    return len(repr(o)) + _offset
  if isinstance(o, (tuple, list, set)):
    for x in o:
      _offset = _type_simplicity_score(x, _offset=_offset + _spacing)
      if _offset > _type_simplicity_limit:
        break
    return _offset
  if isinstance(o, dict):
    for x in o.values():  # ignore keys...
      _offset = _type_simplicity_score(x, _offset=_offset + 10. + _spacing)  # +10 for key
      if _offset > _type_simplicity_limit:
        break
    return _offset
  if isinstance(o, numpy.ndarray):
    _offset += 10.  # prefix/postfix
    if o.size * 2. + _offset > _type_simplicity_limit:  # too big already?
      return o.size * 2. + _offset
    if str(o.dtype).startswith("int"):
      a = _type_simplicity_score(numpy.max(numpy.abs(o))) + _spacing
      return o.size * a + _offset
    a = max([_type_simplicity_score(x) for x in o.flatten()]) + _spacing
    return o.size * a + _offset
  # Unknown object. Fallback > _type_simplicity_limit.
  return _type_simplicity_limit + 1. + _offset
